from dsp_tools.error.xmllib_warnings_util import initialise_warning_file

from .helpers import ListLookup as ListLookup
from .helpers import clean_whitespaces_from_string as clean_whitespaces_from_string
from .helpers import create_footnote_element as create_footnote_element
from .helpers import create_footnote_string as create_footnote_string
from .helpers import create_list_from_string as create_list_from_string
from .helpers import create_non_empty_list_from_string as create_non_empty_list_from_string
from .helpers import create_standoff_link_to_resource as create_standoff_link_to_resource
from .helpers import create_standoff_link_to_uri as create_standoff_link_to_uri
from .helpers import escape_reserved_xml_characters as escape_reserved_xml_characters
from .helpers import find_dates_in_string as find_dates_in_string
from .helpers import find_license_in_string as find_license_in_string
from .helpers import get_list_nodes_from_string_via_list_name as get_list_nodes_from_string_via_list_name
from .helpers import get_list_nodes_from_string_via_property as get_list_nodes_from_string_via_property
from .helpers import make_xsd_compatible_id as make_xsd_compatible_id
from .helpers import make_xsd_compatible_id_with_uuid as make_xsd_compatible_id_with_uuid
from .helpers import reformat_date as reformat_date
from .models.config_options import Calendar as Calendar
from .models.config_options import DateFormat as DateFormat
from .models.config_options import Era as Era
from .models.config_options import NewlineReplacement as NewlineReplacement
from .models.config_options import Permissions as Permissions
from .models.dsp_base_resources import AudioSegmentResource as AudioSegmentResource
from .models.dsp_base_resources import LinkResource as LinkResource
from .models.dsp_base_resources import RegionResource as RegionResource
from .models.dsp_base_resources import VideoSegmentResource as VideoSegmentResource
from .models.licenses.other import LicenseOther as LicenseOther
from .models.licenses.recommended import LicenseRecommended as LicenseRecommended
from .models.res import Resource as Resource
from .models.root import XMLRoot as XMLRoot
from .value_checkers import check_richtext_syntax as check_richtext_syntax
from .value_checkers import is_bool_like as is_bool_like
from .value_checkers import is_color as is_color
from .value_checkers import is_date as is_date
from .value_checkers import is_decimal as is_decimal
from .value_checkers import is_dsp_ark as is_dsp_ark
from .value_checkers import is_dsp_iri as is_dsp_iri
from .value_checkers import is_geoname as is_geoname
from .value_checkers import is_integer as is_integer
from .value_checkers import is_nonempty_value as is_nonempty_value
from .value_checkers import is_timestamp as is_timestamp
from .value_converters import convert_to_bool_string as convert_to_bool_string
from .value_converters import replace_newlines_with_br_tags as replace_newlines_with_br_tags
from .value_converters import replace_newlines_with_paragraph_tags as replace_newlines_with_paragraph_tags
from .value_converters import replace_newlines_with_tags as replace_newlines_with_tags

initialise_warning_file()



from __future__ import annotations

import warnings
from pathlib import Path

from loguru import logger
from lxml import etree

from dsp_tools.commands.xmlupload.prepare_xml_input.iiif_uri_validator import IIIFUriValidator
from dsp_tools.error.custom_warnings import DspToolsUserWarning
from dsp_tools.error.exceptions import InputError


def validate_iiif_uris(root: etree._Element) -> None:
    uris = [uri.strip() for node in root.iter(tag="iiif-uri") if (uri := node.text)]
    if problems := IIIFUriValidator(uris).validate():
        msg = problems.get_msg()
        warnings.warn(DspToolsUserWarning(msg))
        logger.warning(msg)


def check_if_bitstreams_exist(root: etree._Element, imgdir: str) -> None:
    """
    Make sure that all bitstreams referenced in the XML file exist in the imgdir.

    Args:
        root: parsed XML file
        imgdir: folder where the bitstreams are stored

    Raises:
        InputError: if a bitstream does not exist in the imgdir
    """
    multimedia_resources = [x for x in root if any(y.tag == "bitstream" for y in x.iter())]
    for res in multimedia_resources:
        pth = next(Path(x.text.strip()) for x in res.iter() if x.tag == "bitstream" and x.text)
        if not Path(imgdir / pth).is_file():
            raise InputError(
                f"Bitstream '{pth!s}' of resource '{res.attrib['label']}' does not exist in the imgdir '{imgdir}'."
            )



import csv
import inspect
import os
import warnings
from typing import Any
from typing import Never

import regex
from dotenv import load_dotenv

from dsp_tools.error.xmllib_errors import XmllibFileNotFoundError
from dsp_tools.error.xmllib_errors import XmllibInputError
from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings import UserMessageSeverity
from dsp_tools.error.xmllib_warnings import XmllibInputInfo
from dsp_tools.error.xmllib_warnings import XmllibInputWarning
from dsp_tools.utils.ansi_colors import BOLD_YELLOW
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT

load_dotenv()


def initialise_warning_file() -> None:
    if file_path := os.getenv("XMLLIB_WARNINGS_CSV_SAVEPATH"):
        try:
            new_row = ["File", "Severity", "Message", "Resource ID", "Property", "Field"]
            with open(file_path, "w", newline="") as file:
                print(
                    BOLD_YELLOW,
                    f"Warnings generated by the xmllib "
                    f"will be saved to '{file_path}' and will not be printed in the terminal.",
                    RESET_TO_DEFAULT,
                )
                writer = csv.writer(file)
                writer.writerow(new_row)
        except FileNotFoundError:
            raise XmllibFileNotFoundError(
                f"The filepath '{file_path}' you entered in your .env file does not exist. "
                f"Please ensure that the folder you named exists."
            ) from None


def write_message_to_csv(
    file_path: str, msg: MessageInfo, function_trace: str | None, severity: UserMessageSeverity
) -> None:
    new_row = [
        function_trace if function_trace else "",
        str(severity),
        msg.message if msg.message else "",
        msg.resource_id if msg.resource_id else "",
        msg.prop_name if msg.prop_name else "",
        msg.field if msg.field else "",
    ]
    with open(file_path, "a", newline="") as file:
        writer = csv.writer(file)
        writer.writerow(new_row)


def get_user_message_string(msg: MessageInfo, function_trace: str | None) -> str:
    str_list = [f"File '{function_trace}'"] if function_trace else []
    if msg.resource_id:
        str_list.append(f"Resource ID '{msg.resource_id}'")
    if msg.prop_name:
        str_list.append(f"Property '{msg.prop_name}'")
    if msg.field:
        str_list.append(f"Field '{msg.field}'")
    str_list.append(msg.message)
    return " | ".join(str_list)


def _get_calling_code_context() -> str | None:
    """
    Find file name and line number of the file that was written by the user.
    """
    all_stack_frames = inspect.stack()
    frame_files = [x.filename for x in all_stack_frames]
    calling_func_index = _get_stack_frame_number(frame_files)
    if calling_func_index == 0:
        return None
    user_frame_info = all_stack_frames.pop(calling_func_index)
    file_name = user_frame_info.filename.rsplit("/", 1)[1]
    return f"{file_name}:{user_frame_info.lineno}"


def _get_stack_frame_number(file_names: list[str]) -> int:
    """
    Get index number of first python file of the stack trace which is not any more in the dsp-tools code.
    This is the index of the first user python file.
    """
    calling_func_index = -1
    for file in file_names:
        if _filter_stack_frames(file):
            calling_func_index += 1
        else:
            calling_func_index += 1
            break
    return calling_func_index


def _filter_stack_frames(file_path: str) -> bool:
    dsp_tools_path = r"\/dsp[-_]tools\/(xmllib|error)\/"
    if regex.search(dsp_tools_path, file_path):
        return True
    elif regex.search(r"^<[a-zA-Z]+>$", file_path):
        # This is for functions like str(), which appear in the stack trace as filename "<string>"
        return True
    return False


def raise_xmllib_input_error(msg: MessageInfo) -> Never:
    """These are to be used if the error is caused by user input."""
    function_trace = _get_calling_code_context()
    if file_path := os.getenv("XMLLIB_WARNINGS_CSV_SAVEPATH"):
        write_message_to_csv(file_path, msg, function_trace, UserMessageSeverity.ERROR)
    msg_str = get_user_message_string(msg, function_trace)
    raise XmllibInputError(msg_str)


def emit_xmllib_input_warning(msg: MessageInfo) -> None:
    """These are to be used if the error is caused by user input."""
    function_trace = _get_calling_code_context()
    if file_path := os.getenv("XMLLIB_WARNINGS_CSV_SAVEPATH"):
        write_message_to_csv(file_path, msg, function_trace, UserMessageSeverity.WARNING)
    else:
        msg_str = get_user_message_string(msg, function_trace)
        warnings.warn(XmllibInputWarning(msg_str))


def emit_xmllib_input_info(msg: MessageInfo) -> None:
    """These are to be used if the error is caused by user input."""
    function_trace = _get_calling_code_context()
    if file_path := os.getenv("XMLLIB_WARNINGS_CSV_SAVEPATH"):
        write_message_to_csv(file_path, msg, function_trace, UserMessageSeverity.INFO)
    else:
        msg_str = get_user_message_string(msg, function_trace)
        warnings.warn(XmllibInputInfo(msg_str))


def emit_xmllib_input_type_mismatch_warning(
    *,
    expected_type: str,
    value: Any,
    res_id: str | None,
    value_field: str | None = None,
    prop_name: str | None = None,
) -> None:
    """These are to be used if the error is caused by user input."""
    msg_info = MessageInfo(
        message=f"The input should be a valid {expected_type}, your input '{value}' does not match the type.",
        resource_id=res_id,
        prop_name=prop_name,
        field=value_field,
    )
    emit_xmllib_input_warning(msg_info)



# mypy: disable-error-code="method-assign,no-untyped-def"

from uuid import uuid4

from rdflib import RDF
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xmlupload.models.processed.values import ProcessedLink
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.upload_stashed_resptr_props import _make_link_value_create_graph
from dsp_tools.utils.rdflib_constants import KNORA_API

ONTO_STR = "http://0.0.0.0:3333/ontology/9999/onto/v2#"

ONTO = Namespace(ONTO_STR)

RES_IRI_STR = "http://rdfh.ch/9999/res_one"
TARGET_IRI_STR = "http://rdfh.ch/9999/target_resource"
RES_IRI = URIRef(RES_IRI_STR)
RES_TYPE = ONTO.Resource


def test_make_link_value_create_graph():
    link_stash = LinkValueStashItem(
        res_id=RES_IRI_STR,
        res_type=RES_TYPE,
        value=ProcessedLink("target_resource_id", ONTO.hasLink, None, None, str(uuid4())),
    )
    result = _make_link_value_create_graph(link_stash, RES_IRI_STR, TARGET_IRI_STR)
    assert len(result) == 4
    res_type = next(result.objects(RES_IRI, RDF.type))
    assert res_type == RES_TYPE
    val_bn = next(result.objects(RES_IRI, ONTO.hasLinkValue))
    val_type = next(result.objects(val_bn, RDF.type))
    assert val_type == KNORA_API.LinkValue
    target_iri = next(result.objects(val_bn, KNORA_API.linkValueHasTargetIri))
    assert target_iri == URIRef(TARGET_IRI_STR)



# mypy: disable-error-code="no-untyped-def"

from lxml import etree

from dsp_tools.utils.xml_parsing.get_lookups import get_authorship_lookup


def test_get_authorship_lookup():
    root_str = """
    <knora>
        <authorship id="authorship_1">
            <author>
            Lukas Rosenthaler </author>
        </authorship>
        <authorship id="authorship_2">
            <author>\tNora Ammann
            </author>
        </authorship>
        <authorship id="authorship_3">
            <author>Nora    Ammann</author>
            <author>
                Johannes Nussbaum
            </author>
        </authorship>
    </knora>
    """
    root = etree.fromstring(root_str)
    expected = {
        "authorship_1": ["Lukas Rosenthaler"],
        "authorship_2": ["Nora Ammann"],
        "authorship_3": ["Nora Ammann", "Johannes Nussbaum"],
    }
    result = get_authorship_lookup(root)
    assert set(result.keys()) == set(expected.keys())
    for k, v in result.items():
        assert set(v) == set(expected[k])


def test_extract_authorships_from_xml_no_authors():
    root = etree.fromstring("<knora></knora>")
    result = get_authorship_lookup(root)
    assert result == {}



from typing import Any

import pytest

from dsp_tools.commands.project.get import _parse_default_permissions

USER_IRI_PREFIX = "http://www.knora.org/ontology/knora-admin#"
PROJ_IRI = "https://api.dev.dasch.swiss/project/MsOaiQkcQ7-QPxsYBKckfQ"


@pytest.fixture
def private_perms() -> dict[str, Any]:
    perms = [
        {"additionalInformation": f"{USER_IRI_PREFIX}ProjectAdmin", "name": "CR", "permissionCode": 16},
        {"additionalInformation": f"{USER_IRI_PREFIX}ProjectMember", "name": "D", "permissionCode": 8},
    ]
    return {
        "forGroup": f"{USER_IRI_PREFIX}ProjectMember",
        "forProject": PROJ_IRI,
        "hasPermissions": perms,
    }


@pytest.fixture
def public_perms() -> dict[str, Any]:
    perms = [
        {"additionalInformation": f"{USER_IRI_PREFIX}ProjectAdmin", "name": "CR", "permissionCode": 16},
        {"additionalInformation": f"{USER_IRI_PREFIX}ProjectMember", "name": "D", "permissionCode": 8},
        {"additionalInformation": f"{USER_IRI_PREFIX}KnownUser", "name": "V", "permissionCode": 2},
        {"additionalInformation": f"{USER_IRI_PREFIX}UnknownUser", "name": "V", "permissionCode": 2},
    ]
    return {
        "forGroup": f"{USER_IRI_PREFIX}ProjectMember",
        "forProject": PROJ_IRI,
        "hasPermissions": perms,
    }


def test_parse_default_permissions_private(private_perms: dict[str, Any]) -> None:
    assert _parse_default_permissions([private_perms]) == "private"


def test_parse_default_permissions_public(public_perms: dict[str, Any]) -> None:
    assert _parse_default_permissions([public_perms]) == "public"


def test_parse_default_permissions_wrong_target(public_perms: dict[str, Any]) -> None:
    public_perms["forGroup"] = f"{USER_IRI_PREFIX}SystemAdmin"
    assert _parse_default_permissions([public_perms]) == "unknown"


def test_parse_default_permissions_previous_standard(public_perms: dict[str, Any]) -> None:
    public_perms_admin = public_perms.copy()
    public_perms_admin["forGroup"] = f"{USER_IRI_PREFIX}ProjectAdmin"
    assert _parse_default_permissions([public_perms_admin, public_perms]) == "unknown"


def test_parse_default_permissions_with_creator(public_perms: dict[str, Any]) -> None:
    public_perms["hasPermissions"].append(
        {"additionalInformation": f"{USER_IRI_PREFIX}Creator", "name": "CR", "permissionCode": 16}
    )
    assert _parse_default_permissions([public_perms]) == "unknown"



import pytest
from rdflib import RDF
from rdflib import RDFS
from rdflib import SH
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_0_1_cardinality
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_0_n_cardinality
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_1_cardinality
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_1_n_cardinality
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_all_cardinalities
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import _construct_resource_nodeshape
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import construct_cardinality_node_shapes
from dsp_tools.utils.rdflib_constants import API_SHAPES
from dsp_tools.utils.rdflib_constants import DASH
from dsp_tools.utils.rdflib_constants import KNORA_API
from test.unittests.commands.validate_data.constants import ONTO
from test.unittests.commands.validate_data.constants import PREFIXES


class TestCheckTripleNumbersOnto:
    def test_nodeshape(self, onto_for_cardinality: Graph) -> None:
        result = _construct_resource_nodeshape(onto_for_cardinality)
        number_of_resource_classes = 12
        triples_cls_nodeshape = 13 * number_of_resource_classes
        assert len(result) == triples_cls_nodeshape

    def test_one_nodeshape(self) -> None:
        onto = f"""{PREFIXES}
        onto:TestClass a owl:Class ;
            knora-api:canBeInstantiated true ;
            knora-api:isResourceClass true .
        """
        onto_g = Graph()
        onto_g.parse(data=onto)
        result = _construct_resource_nodeshape(onto_g)
        test_cls = ONTO.TestClass
        assert next(result.objects(test_cls, RDF.type)) == SH.NodeShape
        assert next(result.objects(test_cls, DASH.closedByTypes)) == Literal(True)
        properties = list(result.objects(test_cls, SH.property))
        assert len(properties) == 3
        assert next(result.subjects(SH.property, API_SHAPES.hasPermissions_Cardinality)) == test_cls
        label_shape = next(result.subjects(SH.path, RDFS.label))
        assert next(result.subjects(SH.property, label_shape)) == test_cls
        assert next(result.objects(label_shape, RDF.type)) == SH.PropertyShape
        assert next(result.objects(label_shape, SH.minCount)) == Literal(1)
        assert next(result.objects(label_shape, SH.maxCount)) == Literal(1)
        assert next(result.objects(label_shape, SH.severity)) == SH.Violation
        assert next(result.objects(label_shape, SH.message)) == Literal("A label is required")
        standoff_shapes = next(result.subjects(SH.path, KNORA_API.hasStandoffLinkTo))
        assert next(result.subjects(SH.property, standoff_shapes)) == test_cls
        assert next(result.objects(standoff_shapes, RDF.type)) == SH.PropertyShape

    def test_cardinality_1(self, onto_for_cardinality: Graph) -> None:
        result = _construct_1_cardinality(onto_for_cardinality)
        number_of_occurrences_in_onto = 2
        triples_card_1 = 7 * number_of_occurrences_in_onto
        assert len(result) == triples_card_1

    def test_cardinality_0_1(self, onto_for_cardinality: Graph) -> None:
        result = _construct_0_1_cardinality(onto_for_cardinality)
        number_of_occurrences_in_onto = 5
        triples_card_0_1 = 7 * number_of_occurrences_in_onto
        assert len(result) == triples_card_0_1

    def test_cardinality_0_n(self, onto_for_cardinality: Graph) -> None:
        result = _construct_0_n_cardinality(onto_for_cardinality)
        number_of_occurrences_in_onto = 22  # Inheritance included
        triples_card_0_n = 3 * number_of_occurrences_in_onto
        assert len(result) == triples_card_0_n

    def test_cardinality_1_n(self, onto_for_cardinality: Graph) -> None:
        result = _construct_1_n_cardinality(onto_for_cardinality)
        number_of_occurrences_in_onto = 1
        triples_card_1_n = 6 * number_of_occurrences_in_onto
        assert len(result) == triples_card_1_n


def test_construct_cardinality_node_shapes(onto_for_cardinality: Graph) -> None:
    result = construct_cardinality_node_shapes(onto_for_cardinality)
    expected_classes = {
        ONTO.ClassInheritedCardinality,
        ONTO.ClassMixedCard,
        ONTO.CardOneResource,
        ONTO.ClassWithEverything,
        ONTO.TestStillImageRepresentation,
        ONTO.ClassInheritedCardinalityOverwriting,
        ONTO.TestArchiveRepresentation,
        ONTO.TestAudioRepresentation,
        ONTO.TestDocumentRepresentation,
        ONTO.TestMovingImageRepresentation,
        ONTO.TestStillImageRepresentation,
        ONTO.TestTextRepresentation,
        URIRef("http://0.0.0.0:3333/ontology/9999/second-onto/v2#SecondOntoClass"),
    }
    result_classes = set(result.subjects(DASH.closedByTypes, Literal(True)))
    assert result_classes == expected_classes


def test_construct_resource_nodeshape_one_res(one_res_one_prop: Graph) -> None:
    result = _construct_resource_nodeshape(one_res_one_prop)
    subjects = {iri for x in result.triples((None, None, None)) if not isinstance(iri := x[0], BNode)}
    assert len(subjects) == 1
    subject_iri = subjects.pop()
    assert subject_iri == ONTO.CardOneResource
    node_triples = list(result.triples((subject_iri, None, None)))
    num_triples = 5
    assert len(node_triples) == num_triples
    assert next(result.subjects(RDF.type, SH.NodeShape)) == subject_iri
    assert next(result.objects(subject_iri, DASH.closedByTypes)) == Literal(True)
    assert isinstance(next(result.objects(subject_iri, SH.property)), BNode)


def test_construct_resource_nodeshape_no_res(one_bool_prop: Graph) -> None:
    result = _construct_resource_nodeshape(one_bool_prop)
    assert len(result) == 0


class Test1:
    def test_good(self, card_1: Graph) -> None:
        result = _construct_1_cardinality(card_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testBoolean
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 1"

    def test_good_link_value(self, link_prop_card_1: Graph) -> None:
        result = _construct_1_cardinality(link_prop_card_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testHasLinkToCardOneResource
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 1"

    def test_empty_0_1(self, card_0_1: Graph) -> None:
        result = _construct_1_cardinality(card_0_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_1_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_1_cardinality(card_0_n)
        assert len(result) == 0


class Test01:
    def test_good(self, card_0_1: Graph) -> None:
        result = _construct_0_1_cardinality(card_0_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "0"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testDecimalSimpleText
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 0-1"

    def test_good_link_value(self, link_prop_card_01: Graph) -> None:
        result = _construct_0_1_cardinality(link_prop_card_01)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "0"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testHasLinkToCardOneResource
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 0-1"

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_0_1_cardinality(card_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_0_1_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_0_1_cardinality(card_0_n)
        assert len(result) == 0


class Test1N:
    def test_good(self, card_1_n: Graph) -> None:
        result = _construct_1_n_cardinality(card_1_n)
        assert len(result) == 6
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testGeoname
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "Cardinality 1-n"

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_1_n_cardinality(card_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_0_1: Graph) -> None:
        result = _construct_1_n_cardinality(card_0_1)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_1_n_cardinality(card_0_n)
        assert len(result) == 0


class Test0N:
    def test_good(self, card_0_n: Graph) -> None:
        result = _construct_0_n_cardinality(card_0_n)
        assert len(result) == 3
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard
        assert next(result.objects(bn, SH.path)) == ONTO.testSimpleText

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_0_n_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_0_n_cardinality(card_1)
        assert len(result) == 0

    def test_empty_0_1(self, card_0_1: Graph) -> None:
        result = _construct_0_n_cardinality(card_0_1)
        assert len(result) == 0


def test_construct_all_cardinalities(one_res_one_prop: Graph) -> None:
    result = _construct_all_cardinalities(one_res_one_prop)
    assert len(result) == 7
    bn = next(result.subjects(RDF.type, SH.PropertyShape))
    shape_iri = next(result.subjects(SH.property, bn))
    assert shape_iri == ONTO.CardOneResource
    assert str(next(result.objects(bn, SH.minCount))) == "1"
    assert str(next(result.objects(bn, SH.maxCount))) == "1"
    assert next(result.objects(bn, SH.path)) == ONTO.testBoolean
    assert next(result.objects(bn, SH.severity)) == SH.Violation
    assert str(next(result.objects(bn, SH.message))) == "Cardinality 1"


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass
from typing import Any
from typing import cast
from urllib.parse import quote_plus

import requests

from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import OneNode
from dsp_tools.error.exceptions import InternalError
from dsp_tools.utils.request_utils import RequestParameters
from dsp_tools.utils.request_utils import log_request
from dsp_tools.utils.request_utils import log_response


@dataclass
class OntologyClient:
    api_url: str
    shortcode: str

    def get_knora_api(self) -> str:
        url = f"{self.api_url}/ontology/knora-api/v2#"
        headers = {"Accept": "text/turtle"}
        timeout = 60
        log_request(RequestParameters("GET", url, timeout=timeout, headers=headers))
        response = requests.get(url=url, headers=headers, timeout=timeout)
        log_response(response, include_response_content=False)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        return response.text

    def get_ontologies(self) -> tuple[list[str], list[str]]:
        """
        Returns a list of project ontologies as a string in turtle format.
        And a list of the ontology IRIs

        Returns:
            list of ontologies and IRIs
        """
        ontology_iris = self._get_ontology_iris()
        ontologies = [self._get_one_ontology(x) for x in ontology_iris]
        return ontologies, ontology_iris

    def _get_ontology_iris(self) -> list[str]:
        url = f"{self.api_url}/admin/projects/shortcode/{self.shortcode}"
        timeout = 10
        log_request(RequestParameters("GET", url, timeout=timeout))
        response = requests.get(url=url, timeout=timeout)
        log_response(response)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        response_json = cast(dict[str, Any], response.json())
        if not (ontos := response_json.get("project", {}).get("ontologies")):
            raise InternalError(f"The response from the API does not contain any ontologies.\nResponse:{response.text}")
        output = cast(list[str], ontos)
        return output

    def _get_one_ontology(self, ontology_iri: str) -> str:
        url = ontology_iri
        headers = {"Accept": "text/turtle"}
        timeout = 30
        log_request(RequestParameters("GET", url, timeout=timeout, headers=headers))
        response = requests.get(url=url, headers=headers, timeout=timeout)
        log_response(response, include_response_content=False)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        return response.text


@dataclass
class ListClient:
    """Client to request and reformat the lists of a project."""

    api_url: str
    shortcode: str

    def get_lists(self) -> list[OneList]:
        list_json = self._get_all_list_iris()
        all_iris = self._extract_list_iris(list_json)
        all_lists = [self._get_one_list(iri) for iri in all_iris]
        return [self._reformat_one_list(lst) for lst in all_lists]

    def _get_all_list_iris(self) -> dict[str, Any]:
        url = f"{self.api_url}/admin/lists?projectShortcode={self.shortcode}"
        timeout = 10
        log_request(RequestParameters("GET", url, timeout))
        response = requests.get(url=url, timeout=timeout)
        log_response(response)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        json_response = cast(dict[str, Any], response.json())
        return json_response

    def _extract_list_iris(self, response_json: dict[str, Any]) -> list[str]:
        return [x["id"] for x in response_json["lists"]]

    def _get_one_list(self, list_iri: str) -> dict[str, Any]:
        encoded_list_iri = quote_plus(list_iri)
        url = f"{self.api_url}/admin/lists/{encoded_list_iri}"
        timeout = 30
        log_request(RequestParameters("GET", url, timeout))
        response = requests.get(url=url, timeout=timeout)
        log_response(response, include_response_content=False)
        if not response.ok:
            raise InternalError(f"Failed Request: {response.status_code} {response.text}")
        response_json = cast(dict[str, Any], response.json())
        return response_json

    def _reformat_one_list(self, response_json: dict[str, Any]) -> OneList:
        list_name = response_json["list"]["listinfo"]["name"]
        list_id = response_json["list"]["listinfo"]["id"]
        nodes = response_json["list"]["children"]
        all_nodes = []
        for child in nodes:
            all_nodes.append(OneNode(child["name"], child["id"]))
            if node_child := child.get("children"):
                self._reformat_children(node_child, all_nodes)
        return OneList(list_iri=list_id, list_name=list_name, nodes=all_nodes)

    def _reformat_children(self, list_child: list[dict[str, Any]], current_nodes: list[OneNode]) -> None:
        for child in list_child:
            current_nodes.append(OneNode(child["name"], child["id"]))
            if grand_child := child.get("children"):
                self._reformat_children(grand_child, current_nodes)



from typing import cast
from uuid import uuid4

from loguru import logger

from dsp_tools.commands.xmlupload.models.lookup_models import XmlReferenceLookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.processed.file_values import ProcessedFileMetadata
from dsp_tools.commands.xmlupload.models.processed.file_values import ProcessedFileValue
from dsp_tools.commands.xmlupload.models.processed.file_values import ProcessedIIIFUri
from dsp_tools.commands.xmlupload.models.processed.res import MigrationMetadata
from dsp_tools.commands.xmlupload.models.processed.res import ProcessedResource
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedBoolean
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedColor
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedDate
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedDecimal
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedGeometry
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedGeoname
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedInt
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedInterval
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedLink
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedList
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedRichtext
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedSimpleText
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedTime
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedUri
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedValue
from dsp_tools.commands.xmlupload.prepare_xml_input.ark2iri import convert_ark_v0_to_resource_iri
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import TypeTransformerMapper
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import assert_is_string
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import assert_is_tuple
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_boolean
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_date
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_decimal
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_geometry
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_integer
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_interval
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_richtext
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_simpletext
from dsp_tools.error.exceptions import XmlUploadAuthorshipsNotFoundError
from dsp_tools.error.exceptions import XmlUploadListNodeNotFoundError
from dsp_tools.error.exceptions import XmlUploadPermissionsNotFoundError
from dsp_tools.legacy_models.datetimestamp import DateTimeStamp
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValueMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedMigrationMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedValue

TYPE_TRANSFORMER_MAPPER: dict[KnoraValueType, TypeTransformerMapper] = {
    KnoraValueType.BOOLEAN_VALUE: TypeTransformerMapper(ProcessedBoolean, transform_boolean),
    KnoraValueType.COLOR_VALUE: TypeTransformerMapper(ProcessedColor, assert_is_string),
    KnoraValueType.DECIMAL_VALUE: TypeTransformerMapper(ProcessedDecimal, transform_decimal),
    KnoraValueType.DATE_VALUE: TypeTransformerMapper(ProcessedDate, transform_date),
    KnoraValueType.GEOM_VALUE: TypeTransformerMapper(ProcessedGeometry, transform_geometry),
    KnoraValueType.GEONAME_VALUE: TypeTransformerMapper(ProcessedGeoname, assert_is_string),
    KnoraValueType.INT_VALUE: TypeTransformerMapper(ProcessedInt, transform_integer),
    KnoraValueType.INTERVAL_VALUE: TypeTransformerMapper(ProcessedInterval, transform_interval),
    KnoraValueType.TIME_VALUE: TypeTransformerMapper(ProcessedTime, assert_is_string),
    KnoraValueType.SIMPLETEXT_VALUE: TypeTransformerMapper(ProcessedSimpleText, transform_simpletext),
    KnoraValueType.URI_VALUE: TypeTransformerMapper(ProcessedUri, assert_is_string),
}


def get_processed_resources(
    resources: list[ParsedResource], lookups: XmlReferenceLookups, is_on_prod_like_server: bool
) -> list[ProcessedResource]:
    logger.debug("Transform ParsedResource into ProcessedResource")
    return [_get_one_resource(res, lookups, is_on_prod_like_server) for res in resources]


def _get_one_resource(
    resource: ParsedResource, lookups: XmlReferenceLookups, is_on_prod_like_server: bool
) -> ProcessedResource:
    permissions = _resolve_permission(resource.permissions_id, lookups.permissions)
    values = [_get_one_processed_value(val, lookups) for val in resource.values]
    migration_metadata = None
    file_val, iiif_uri = _resolve_file_value(resource, lookups, is_on_prod_like_server)
    if resource.migration_metadata:
        migration_metadata = _get_resource_migration_metadata(resource.migration_metadata)
    return ProcessedResource(
        res_id=resource.res_id,
        type_iri=resource.res_type,
        label=resource.label,
        permissions=permissions,
        values=values,
        file_value=file_val,
        iiif_uri=iiif_uri,
        migration_metadata=migration_metadata,
    )


def _get_resource_migration_metadata(metadata: ParsedMigrationMetadata) -> MigrationMetadata:
    res_iri = metadata.iri
    # ARK takes precedence over the IRI,
    # but must be transformed into an IRI as it is only for external reference and not consistent with a DB IRI
    if metadata.ark:
        res_iri = convert_ark_v0_to_resource_iri(metadata.ark)
    date = DateTimeStamp(metadata.creation_date) if metadata.creation_date else None
    return MigrationMetadata(res_iri, date)


def _resolve_file_value(
    resource: ParsedResource, lookups: XmlReferenceLookups, is_on_prod_like_server: bool
) -> tuple[None | ProcessedFileValue, None | ProcessedIIIFUri]:
    file_val, iiif_uri = None, None
    if not resource.file_value:
        return file_val, iiif_uri

    if is_on_prod_like_server:
        metadata = _get_file_metadata(resource.file_value.metadata, lookups)
    else:
        metadata = _get_file_metadata_for_test_environments(resource.file_value.metadata, lookups)
    if resource.file_value.value_type == KnoraValueType.STILL_IMAGE_IIIF:
        iiif_uri = _get_iiif_uri_value(resource.file_value, metadata)
    else:
        file_val = _get_file_value(
            val=resource.file_value, metadata=metadata, res_id=resource.res_id, res_label=resource.label
        )
    return file_val, iiif_uri


def _get_file_value(
    val: ParsedFileValue, metadata: ProcessedFileMetadata, res_id: str, res_label: str
) -> ProcessedFileValue:
    file_type = cast(KnoraValueType, val.value_type)
    file_val = assert_is_string(val.value)
    return ProcessedFileValue(
        value=file_val,
        file_type=file_type,
        metadata=metadata,
        res_id=res_id,
        res_label=res_label,
    )


def _get_iiif_uri_value(iiif_uri: ParsedFileValue, metadata: ProcessedFileMetadata) -> ProcessedIIIFUri:
    file_val = assert_is_string(iiif_uri.value)
    return ProcessedIIIFUri(file_val, metadata)


def _get_file_metadata(file_metadata: ParsedFileValueMetadata, lookups: XmlReferenceLookups) -> ProcessedFileMetadata:
    license_iri = assert_is_string(file_metadata.license_iri)
    copyright_holder = assert_is_string(file_metadata.copyright_holder)
    auth_id = assert_is_string(file_metadata.authorship_id)
    authorships = _resolve_authorship(auth_id, lookups.authorships)
    permissions = _resolve_permission(file_metadata.permissions_id, lookups.permissions)
    return ProcessedFileMetadata(
        license_iri=license_iri,
        copyright_holder=copyright_holder,
        authorships=authorships,
        permissions=permissions,
    )


def _get_file_metadata_for_test_environments(
    metadata: ParsedFileValueMetadata, lookups: XmlReferenceLookups
) -> ProcessedFileMetadata:
    lic_iri = metadata.license_iri or "http://rdfh.ch/licenses/unknown"
    copy_right = metadata.copyright_holder if metadata.copyright_holder else "DUMMY"
    if not metadata.authorship_id:
        authorship = ["DUMMY"]
    else:
        authorship = _resolve_authorship(metadata.authorship_id, lookups.authorships)
    permissions = _resolve_permission(metadata.permissions_id, lookups.permissions)
    return ProcessedFileMetadata(
        license_iri=lic_iri,
        copyright_holder=copy_right,
        authorships=authorship,
        permissions=permissions,
    )


def _resolve_authorship(authorship_id: str, lookup: dict[str, list[str]]) -> list[str]:
    if not (found := lookup.get(authorship_id)):
        raise XmlUploadAuthorshipsNotFoundError(f"Could not find authorships for value: {authorship_id}")
    return found


def _get_one_processed_value(val: ParsedValue, lookups: XmlReferenceLookups) -> ProcessedValue:
    match val.value_type:
        case KnoraValueType.LIST_VALUE:
            return _get_list_value(val, lookups)
        case KnoraValueType.LINK_VALUE:
            return _get_link_value(val, lookups)
        case KnoraValueType.RICHTEXT_VALUE:
            return _get_richtext_value(val, lookups)
        case _ as val_type:
            transformation_mapper = TYPE_TRANSFORMER_MAPPER[val_type]
            return _get_generic_value(val=val, lookups=lookups, transformation_mapper=transformation_mapper)


def _get_generic_value(
    val: ParsedValue, lookups: XmlReferenceLookups, transformation_mapper: TypeTransformerMapper
) -> ProcessedValue:
    transformed_value = transformation_mapper.val_transformer(val.value)
    permission_val = _resolve_permission(val.permissions_id, lookups.permissions)
    return transformation_mapper.val_type(transformed_value, val.prop_name, val.comment, permission_val)


def _get_link_value(val: ParsedValue, lookups: XmlReferenceLookups) -> ProcessedValue:
    transformed_value = assert_is_string(val.value)
    permission_val = _resolve_permission(val.permissions_id, lookups.permissions)
    link_val: ProcessedValue = ProcessedLink(
        value=transformed_value,
        prop_iri=val.prop_name,
        comment=val.comment,
        permissions=permission_val,
        value_uuid=str(uuid4()),
    )
    return link_val


def _get_list_value(val: ParsedValue, lookups: XmlReferenceLookups) -> ProcessedValue:
    tuple_val = assert_is_tuple(val.value)
    if not (list_iri := lookups.listnodes.get(tuple_val)):
        raise XmlUploadListNodeNotFoundError(f"Could not find list IRI for value: {tuple_val}")
    permission_val = _resolve_permission(val.permissions_id, lookups.permissions)
    list_val: ProcessedValue = ProcessedList(
        value=list_iri,
        prop_iri=val.prop_name,
        comment=val.comment,
        permissions=permission_val,
    )
    return list_val


def _get_richtext_value(val: ParsedValue, lookups: XmlReferenceLookups) -> ProcessedValue:
    transformed_value = transform_richtext(val.value)
    permission_val = _resolve_permission(val.permissions_id, lookups.permissions)
    richtext: ProcessedValue = ProcessedRichtext(
        value=transformed_value,
        prop_iri=val.prop_name,
        comment=val.comment,
        permissions=permission_val,
        resource_references=transformed_value.find_internal_ids(),
        value_uuid=str(uuid4()),
    )
    return richtext


def _resolve_permission(permissions: str | None, permissions_lookup: dict[str, Permissions]) -> Permissions | None:
    """Resolve the permission into a string that can be sent to the API."""
    if permissions:
        if not (per := permissions_lookup.get(permissions)):
            raise XmlUploadPermissionsNotFoundError(f"Could not find permissions for value: {permissions}")
        return per
    return None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import Protocol

from dsp_tools.error.xmllib_errors import XmllibInputError
from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_type_mismatch_warning
from dsp_tools.error.xmllib_warnings_util import raise_xmllib_input_error
from dsp_tools.utils.data_formats.uri_util import is_uri
from dsp_tools.xmllib.internal.checkers import check_and_inform_about_angular_brackets
from dsp_tools.xmllib.internal.circumvent_circular_imports import parse_richtext_as_xml
from dsp_tools.xmllib.internal.input_converters import check_and_fix_is_non_empty_string
from dsp_tools.xmllib.internal.input_converters import check_and_get_corrected_comment
from dsp_tools.xmllib.models.config_options import NewlineReplacement
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.value_checkers import is_color
from dsp_tools.xmllib.value_checkers import is_date
from dsp_tools.xmllib.value_checkers import is_decimal
from dsp_tools.xmllib.value_checkers import is_geoname
from dsp_tools.xmllib.value_checkers import is_integer
from dsp_tools.xmllib.value_checkers import is_link_value
from dsp_tools.xmllib.value_checkers import is_nonempty_value
from dsp_tools.xmllib.value_checkers import is_timestamp
from dsp_tools.xmllib.value_converters import convert_to_bool_string
from dsp_tools.xmllib.value_converters import replace_newlines_with_tags


class Value(Protocol):
    value: str
    prop_name: str
    permissions: Permissions
    comment: str | None


@dataclass
class BooleanValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> BooleanValue:
        try:
            val = str(convert_to_bool_string(value)).lower()
        except XmllibInputError:
            emit_xmllib_input_type_mismatch_warning(
                expected_type="bool", value=value, res_id=resource_id, prop_name=prop_name
            )
            val = str(value)
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=val, prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class ColorValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> ColorValue:
        if not is_color(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="color", value=value, res_id=resource_id, prop_name=prop_name
            )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=str(value), prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class DateValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> DateValue:
        if not is_date(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="date", value=value, res_id=resource_id, prop_name=prop_name
            )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=str(value), prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class DecimalValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> DecimalValue:
        if not is_decimal(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="decimal", value=value, res_id=resource_id, prop_name=prop_name
            )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=str(value), prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class GeonameValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> GeonameValue:
        if not is_geoname(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="geoname", value=value, res_id=resource_id, prop_name=prop_name
            )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=str(value), prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class IntValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> IntValue:
        if not is_integer(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="integer", value=value, res_id=resource_id, prop_name=prop_name
            )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=str(value), prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class LinkValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> LinkValue:
        if not is_link_value(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="xsd:ID or DSP resource IRI", value=value, res_id=resource_id, prop_name=prop_name
            )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=str(value), prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class ListValue(Value):
    value: str
    list_name: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        list_name: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> ListValue:
        if str(value).startswith("http://rdfh.ch/lists/"):
            list_str = ""
        else:
            list_str = str(list_name)
            if not is_nonempty_value(list_name):
                emit_xmllib_input_type_mismatch_warning(
                    expected_type="list name", value=list_name, res_id=resource_id, prop_name=prop_name
                )
            if not is_nonempty_value(value):
                emit_xmllib_input_type_mismatch_warning(
                    expected_type="list node", value=value, res_id=resource_id, prop_name=prop_name
                )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(
            value=str(value),
            list_name=list_str,
            prop_name=prop_name,
            permissions=permissions,
            comment=fixed_comment,
        )


@dataclass
class SimpleText(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> SimpleText:
        converted_val = check_and_fix_is_non_empty_string(value=value, res_id=resource_id, prop_name=prop_name)
        check_and_inform_about_angular_brackets(value=value, res_id=resource_id, prop_name=prop_name)
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=converted_val, prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class Richtext(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
        newline_replacement: NewlineReplacement = NewlineReplacement.NONE,
    ) -> Richtext:
        converted_val = check_and_fix_is_non_empty_string(value=value, res_id=resource_id, prop_name=prop_name)
        converted_val = replace_newlines_with_tags(converted_val, newline_replacement)
        result = parse_richtext_as_xml(converted_val)
        if isinstance(result, MessageInfo):
            raise_xmllib_input_error(result)
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=converted_val, prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class TimeValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> TimeValue:
        if not is_timestamp(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="timestamp", value=value, res_id=resource_id, prop_name=prop_name
            )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=str(value), prop_name=prop_name, permissions=permissions, comment=fixed_comment)


@dataclass
class UriValue(Value):
    value: str
    prop_name: str
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    comment: str | None = None

    @classmethod
    def new(
        cls,
        value: Any,
        prop_name: str,
        permissions: Permissions,
        comment: str | None,
        resource_id: str | None,
    ) -> UriValue:
        if not is_uri(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="uri", value=value, res_id=resource_id, prop_name=prop_name
            )
        fixed_comment = check_and_get_corrected_comment(comment, resource_id, prop_name)
        return cls(value=str(value), prop_name=prop_name, permissions=permissions, comment=fixed_comment)



from __future__ import annotations

from typing import cast

import pandas as pd
import pytest
import regex

from dsp_tools.commands.excel2json.json_header import _check_all_users
from dsp_tools.commands.excel2json.json_header import _check_descriptions
from dsp_tools.commands.excel2json.json_header import _check_if_sheets_are_filled_and_exist
from dsp_tools.commands.excel2json.json_header import _check_keywords
from dsp_tools.commands.excel2json.json_header import _check_one_user
from dsp_tools.commands.excel2json.json_header import _check_prefixes
from dsp_tools.commands.excel2json.json_header import _check_project_sheet
from dsp_tools.commands.excel2json.json_header import _do_all_checks
from dsp_tools.commands.excel2json.json_header import _extract_descriptions
from dsp_tools.commands.excel2json.json_header import _extract_keywords
from dsp_tools.commands.excel2json.json_header import _extract_one_user
from dsp_tools.commands.excel2json.json_header import _extract_prefixes
from dsp_tools.commands.excel2json.json_header import _extract_project
from dsp_tools.commands.excel2json.json_header import _extract_users
from dsp_tools.commands.excel2json.json_header import _process_file
from dsp_tools.commands.excel2json.models.input_error import AtLeastOneValueRequiredProblem
from dsp_tools.commands.excel2json.models.input_error import EmptySheetsProblem
from dsp_tools.commands.excel2json.models.input_error import ExcelFileProblem
from dsp_tools.commands.excel2json.models.input_error import ExcelSheetProblem
from dsp_tools.commands.excel2json.models.input_error import InvalidExcelContentProblem
from dsp_tools.commands.excel2json.models.input_error import MissingValuesProblem
from dsp_tools.commands.excel2json.models.input_error import MoreThanOneRowProblem
from dsp_tools.commands.excel2json.models.input_error import PositionInExcel
from dsp_tools.commands.excel2json.models.input_error import RequiredColumnMissingProblem
from dsp_tools.commands.excel2json.models.json_header import Descriptions
from dsp_tools.commands.excel2json.models.json_header import Keywords
from dsp_tools.commands.excel2json.models.json_header import Prefixes
from dsp_tools.commands.excel2json.models.json_header import Project
from dsp_tools.commands.excel2json.models.json_header import User
from dsp_tools.commands.excel2json.models.json_header import Users
from dsp_tools.error.custom_warnings import DspToolsFutureWarning


class TestCheckAll:
    def test_good(
        self,
        prefixes_good: pd.DataFrame,
        project_good_no_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
        licenses_good: pd.DataFrame,
        users_good: pd.DataFrame,
    ) -> None:
        test_dict = {
            "prefixes": prefixes_good,
            "project": project_good_no_zero,
            "description": description_good,
            "keywords": keywords_good,
            "licenses": licenses_good,
            "users": users_good,
        }
        assert not _do_all_checks(test_dict)

    def test_warns_licenses_missing(
        self,
        prefixes_good: pd.DataFrame,
        project_good_no_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
        users_good: pd.DataFrame,
    ) -> None:
        test_dict = {
            "prefixes": prefixes_good,
            "project": project_good_no_zero,
            "description": description_good,
            "keywords": keywords_good,
            "users": users_good,
        }
        expected_msg = regex.escape(
            "The json_header.xlsx file does not have a sheet containing the enabled licenses. "
            "Please note that this will become mandatory in the future."
        )
        with pytest.warns(DspToolsFutureWarning, match=expected_msg):
            checks = _do_all_checks(test_dict)
        assert not checks

    def test_empty_licenses(
        self,
        prefixes_good: pd.DataFrame,
        project_good_no_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
        users_good: pd.DataFrame,
    ) -> None:
        test_dict = {
            "prefixes": prefixes_good,
            "project": project_good_no_zero,
            "description": description_good,
            "keywords": keywords_good,
            "licenses": pd.DataFrame({}),
            "users": users_good,
        }
        result = _do_all_checks(test_dict)
        assert isinstance(result, ExcelFileProblem)
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, ExcelSheetProblem)
        assert problem.sheet_name == "licenses"
        assert len(problem.problems) == 1
        assert isinstance(problem.problems[0], RequiredColumnMissingProblem)
        assert problem.problems[0].columns == ["enabled"]

    def test_empty_sheet_and_missing_sheet(self) -> None:
        test_dict = {
            "prefixes": pd.DataFrame({}),
            "project": pd.DataFrame({"one": [1]}),
            "description": pd.DataFrame({"one": [1]}),
        }
        with pytest.warns(
            DspToolsFutureWarning,
            match=regex.escape("The json_header.xlsx file does not have a sheet containing the enabled licenses"),
        ):
            result = _do_all_checks(test_dict)
        assert isinstance(result, ExcelFileProblem)
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, EmptySheetsProblem)
        assert set(problem.sheet_names) == {"keywords", "prefixes"}

    def test_user_problems(
        self,
        prefixes_good: pd.DataFrame,
        project_good_no_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
        users_wrong_lang: pd.DataFrame,
    ) -> None:
        test_dict = {
            "prefixes": prefixes_good,
            "project": project_good_no_zero,
            "description": description_good,
            "keywords": keywords_good,
            "users": users_wrong_lang,
        }
        with pytest.warns(
            DspToolsFutureWarning,
            match=regex.escape("The json_header.xlsx file does not have a sheet containing the enabled licenses"),
        ):
            result = _do_all_checks(test_dict)
        assert isinstance(result, ExcelFileProblem)
        assert len(result.problems) == 1
        sheet_problem = result.problems[0]
        assert isinstance(sheet_problem, ExcelSheetProblem)
        assert sheet_problem.sheet_name == "users"
        assert len(sheet_problem.problems) == 1
        user_problem = sheet_problem.problems[0]
        assert isinstance(user_problem, InvalidExcelContentProblem)
        assert user_problem.expected_content == "One of: en, de, fr, it, rm"
        assert user_problem.actual_content == "other"
        assert user_problem.excel_position.column == "lang"
        assert user_problem.excel_position.row == 2


class TestCheckIfSheetsExist:
    def test_good_with_users(self) -> None:
        test_dict = {
            "prefixes": pd.DataFrame({"one": [1]}),
            "project": pd.DataFrame({"one": [1]}),
            "description": pd.DataFrame({"one": [1]}),
            "keywords": pd.DataFrame({"one": [1]}),
            "users": pd.DataFrame({"one": [1]}),
        }
        assert not _check_if_sheets_are_filled_and_exist(test_dict)

    def test_good_no_users(self) -> None:
        test_dict = {
            "prefixes": pd.DataFrame({"one": [1]}),
            "project": pd.DataFrame({"one": [1]}),
            "description": pd.DataFrame({"one": [1]}),
            "keywords": pd.DataFrame({"one": [1]}),
        }
        assert not _check_if_sheets_are_filled_and_exist(test_dict)

    def test_missing_sheet(self) -> None:
        test_dict = {
            "prefixes": pd.DataFrame({"one": [1]}),
            "project": pd.DataFrame({"one": [1]}),
            "keywords": pd.DataFrame({"one": [1]}),
            "users": pd.DataFrame({"one": [1]}),
        }
        result = _check_if_sheets_are_filled_and_exist(test_dict)
        assert isinstance(result, ExcelFileProblem)
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, EmptySheetsProblem)
        assert problem.sheet_names == ["description"]

    def test_empty_sheet(self) -> None:
        test_dict = {
            "prefixes": pd.DataFrame({"one": [1]}),
            "project": pd.DataFrame({"one": [1]}),
            "description": pd.DataFrame({"one": [1]}),
            "keywords": pd.DataFrame({}),
        }
        result = _check_if_sheets_are_filled_and_exist(test_dict)
        assert isinstance(result, ExcelFileProblem)
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, EmptySheetsProblem)
        assert problem.sheet_names == ["keywords"]


class TestCheckPrefix:
    def test_good(self, prefixes_good: pd.DataFrame) -> None:
        assert not _check_prefixes(prefixes_good)

    def test_missing_value(self, prefixes_missing_val: pd.DataFrame) -> None:
        result = _check_prefixes(prefixes_missing_val)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "prefixes"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, MissingValuesProblem)
        assert len(problem.locations) == 1
        location = problem.locations[0]
        assert location.column == "prefixes"
        assert location.row == 3

    def test_missing_column(self, prefixes_missing_col: pd.DataFrame) -> None:
        result = _check_prefixes(prefixes_missing_col)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "prefixes"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, RequiredColumnMissingProblem)
        assert problem.columns == ["prefixes"]

    def test_wrong_value(self, prefixes_wrong_val: pd.DataFrame) -> None:
        result = _check_prefixes(prefixes_wrong_val)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "prefixes"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, InvalidExcelContentProblem)
        assert problem.expected_content == "A valid URI"
        assert problem.actual_content == "not a uri"
        assert problem.excel_position.column == "uri"
        assert problem.excel_position.row == 3


class TestCheckProject:
    def test_good(self, project_good_no_zero: pd.DataFrame) -> None:
        assert not _check_project_sheet(project_good_no_zero)

    def test_missing_col(self, project_missing_col: pd.DataFrame) -> None:
        result = _check_project_sheet(project_missing_col)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "project"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, RequiredColumnMissingProblem)
        assert problem.columns == ["shortcode"]

    def test_more_than_one_row(self, project_too_many_rows: pd.DataFrame) -> None:
        result = _check_project_sheet(project_too_many_rows)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "project"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, MoreThanOneRowProblem)
        assert problem.num_rows == 2

    def test_missing_values(self, project_missing_val: pd.DataFrame) -> None:
        result = _check_project_sheet(project_missing_val)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "project"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, MissingValuesProblem)
        assert len(problem.locations) == 1
        location = problem.locations[0]
        assert isinstance(location, PositionInExcel)
        assert location.column == "shortcode"
        assert location.row == 2

    def test_invalid_permissions(self, project_invalid_permissions: pd.DataFrame) -> None:
        result = _check_project_sheet(project_invalid_permissions)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "project"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, InvalidExcelContentProblem)
        assert problem.expected_content == "One of: public, private"
        assert problem.actual_content == "invalid"
        assert problem.excel_position.column == "default_permissions"
        assert problem.excel_position.row == 2


class TestCheckDescription:
    def test_good(self, description_good: pd.DataFrame) -> None:
        assert not _check_descriptions(description_good)

    def test_more_than_one_row(self, description_too_many_rows: pd.DataFrame) -> None:
        result = _check_descriptions(description_too_many_rows)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "description"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, MoreThanOneRowProblem)
        assert problem.num_rows == 2

    def test_missing_values(self, description_missing_val: pd.DataFrame) -> None:
        result = _check_descriptions(description_missing_val)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "description"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, AtLeastOneValueRequiredProblem)
        expected_cols = {"description_en", "description_de", "description_fr", "description_it", "description_rm"}
        assert set(problem.columns) == expected_cols
        assert problem.row_num == 2

    def test_missing_col(self, description_missing_col: pd.DataFrame) -> None:
        result = _check_descriptions(description_missing_col)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "description"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, AtLeastOneValueRequiredProblem)
        expected_cols = {"description_en", "description_de", "description_fr", "description_it", "description_rm"}
        assert set(problem.columns) == expected_cols
        assert problem.row_num == 2


class TestCheckKeywords:
    def test_good(self, keywords_good: pd.DataFrame) -> None:
        assert not _check_keywords(keywords_good)

    def test_missing_column(self, keywords_missing_col: pd.DataFrame) -> None:
        result = _check_keywords(keywords_missing_col)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "keywords"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, RequiredColumnMissingProblem)
        assert problem.columns == ["keywords"]

    def test_missing_value(self, keywords_missing_val: pd.DataFrame) -> None:
        result = _check_keywords(keywords_missing_val)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "keywords"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, MissingValuesProblem)
        assert len(problem.locations) == 1
        location = problem.locations[0]
        assert location.column == "keywords"
        assert not location.row


class TestCheckUsers:
    def test_good(self, users_good: pd.DataFrame) -> None:
        assert not _check_all_users(users_good)

    def test_missing_column(self, users_missing_col: pd.DataFrame) -> None:
        result = _check_all_users(users_missing_col)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "users"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, RequiredColumnMissingProblem)
        assert problem.columns == ["password"]

    def test_missing_value(self, users_missing_val: pd.DataFrame) -> None:
        result = _check_all_users(users_missing_val)
        assert isinstance(result, ExcelSheetProblem)
        assert result.sheet_name == "users"
        assert len(result.problems) == 1
        problem = result.problems[0]
        assert isinstance(problem, MissingValuesProblem)
        assert len(problem.locations) == 1
        location = problem.locations[0]
        assert location.column == "username"
        assert location.row == 2


class TestCheckOneUser:
    def test_good(self, user_good: pd.Series[str]) -> None:
        assert not _check_one_user(user_good, 2)

    def test_bad_lang(self, user_wrong_lang: pd.Series[str]) -> None:
        result = _check_one_user(user_wrong_lang, 2)
        assert len(result) == 1
        problem = result[0]
        assert isinstance(problem, InvalidExcelContentProblem)
        assert problem.expected_content == "One of: en, de, fr, it, rm"
        assert problem.actual_content == "other"
        assert problem.excel_position.column == "lang"
        assert problem.excel_position.row == 2

    def test_bad_email(self, user_wrong_email: pd.Series[str]) -> None:
        result = _check_one_user(user_wrong_email, 2)
        assert len(result) == 1
        problem = result[0]
        assert isinstance(problem, InvalidExcelContentProblem)
        assert problem.expected_content == "A valid email adress"
        assert problem.actual_content == "not an email"
        assert problem.excel_position.column == "email"
        assert problem.excel_position.row == 2

    def test_bad_role(self, user_wrong_role: pd.Series[str]) -> None:
        result = _check_one_user(user_wrong_role, 2)
        assert len(result) == 1
        problem = result[0]
        assert isinstance(problem, InvalidExcelContentProblem)
        assert problem.expected_content == "One of: projectadmin, projectmember"
        assert problem.actual_content == "other"
        assert problem.excel_position.column == "role"
        assert problem.excel_position.row == 2


class TestProcessFile:
    def test_no_user(
        self,
        prefixes_good: pd.DataFrame,
        project_good_no_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
    ) -> None:
        test_dict = {
            "prefixes": prefixes_good,
            "project": project_good_no_zero,
            "description": description_good,
            "keywords": keywords_good,
        }
        result = _process_file(test_dict)
        assert isinstance(result.prefixes, Prefixes)
        assert isinstance(result.project, Project)

    def test_with_user(
        self,
        prefixes_good: pd.DataFrame,
        project_good_missing_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
        users_good: pd.DataFrame,
    ) -> None:
        test_dict = {
            "prefixes": prefixes_good,
            "project": project_good_missing_zero,
            "description": description_good,
            "keywords": keywords_good,
            "users": users_good,
        }
        result = _process_file(test_dict)
        assert isinstance(result.prefixes, Prefixes)
        assert isinstance(result.project, Project)


def test_extract_prefixes(prefixes_good: pd.DataFrame) -> None:
    result = _extract_prefixes(prefixes_good)
    assert isinstance(result, Prefixes)
    assert result.prefixes == {"foaf": "http://xmlns.com/foaf/0.1/", "sdh": "https://ontome.net/ns/sdhss/"}


class TestExtractProject:
    def test_with_users(
        self,
        project_good_missing_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
        users_good: pd.DataFrame,
        licenses_good: pd.DataFrame,
    ) -> None:
        test_dict = {
            "project": project_good_missing_zero,
            "description": description_good,
            "keywords": keywords_good,
            "licenses": licenses_good,
            "users": users_good,
        }
        result = _extract_project(test_dict)
        assert isinstance(result, Project)
        assert result.shortcode == "0011"
        assert result.shortname == "name"
        assert result.longname == "long"
        assert isinstance(result.descriptions, Descriptions)
        assert isinstance(result.keywords, Keywords)
        assert len(result.licenses.licenses) == 2
        assert isinstance(result.users, Users)

    def test_no_users(
        self,
        prefixes_good: pd.DataFrame,
        project_good_no_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
        licenses_good: pd.DataFrame,
    ) -> None:
        test_dict = {
            "prefixes": prefixes_good,
            "project": project_good_no_zero,
            "description": description_good,
            "keywords": keywords_good,
            "licenses": licenses_good,
        }
        result = _extract_project(test_dict)
        assert result.shortcode == "1111"
        assert result.shortname == "name"
        assert result.longname == "long"
        assert isinstance(result.descriptions, Descriptions)
        assert isinstance(result.keywords, Keywords)
        assert len(result.licenses.licenses) == 2
        assert not result.users

    def test_no_licenses(
        self,
        prefixes_good: pd.DataFrame,
        project_good_no_zero: pd.DataFrame,
        description_good: pd.DataFrame,
        keywords_good: pd.DataFrame,
    ) -> None:
        test_dict = {
            "prefixes": prefixes_good,
            "project": project_good_no_zero,
            "description": description_good,
            "keywords": keywords_good,
        }
        result = _extract_project(test_dict)
        assert result.shortcode == "1111"
        assert result.shortname == "name"
        assert result.longname == "long"
        assert isinstance(result.descriptions, Descriptions)
        assert isinstance(result.keywords, Keywords)
        assert len(result.licenses.licenses) == 0
        assert not result.users


def test_extract_descriptions(description_good: pd.DataFrame) -> None:
    result = _extract_descriptions(description_good)
    assert isinstance(result, Descriptions)
    assert result.descriptions == {"en": "english", "fr": "french"}


def test_extract_keywords(keywords_good: pd.DataFrame) -> None:
    result = _extract_keywords(keywords_good)
    assert isinstance(result, Keywords)
    assert set(result.keywords) == {"one", "three"}


class TestUsers:
    def test_extract_users(self, users_good: pd.DataFrame) -> None:
        result = _extract_users(users_good)
        assert isinstance(result, Users)
        assert len(result.users) == 3

    def test_projectadmin(self, users_good: pd.DataFrame) -> None:
        str_row = cast("pd.Series[str]", users_good.loc[1, :])
        result = _extract_one_user(str_row)
        assert isinstance(result, User)
        assert result.username == "Caterpillar"
        assert result.email == "caterpillar@dasch.swiss"
        assert result.givenName == "Caterpillar"
        assert result.familyName == "Wonderland"
        assert result.password == "alice7652"
        assert result.lang == "de"
        assert result.isProjectAdmin

    def test_projectmember(self, users_good: pd.DataFrame) -> None:
        str_row = cast("pd.Series[str]", users_good.loc[2, :])
        result = _extract_one_user(str_row)
        assert isinstance(result, User)
        assert result.username == "WhiteRabbit"
        assert result.email == "white.rabbit@dasch.swiss"
        assert result.givenName == "White"
        assert result.familyName == "Rabbit"
        assert result.password == "alice8711"
        assert result.lang == "fr"
        assert not result.isProjectAdmin


if __name__ == "__main__":
    pytest.main([__file__])



import importlib.resources
from pathlib import Path

from loguru import logger
from rdflib import Graph

from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.legal_info_client_live import LegalInfoClientLive
from dsp_tools.commands.validate_data.api_clients import ListClient
from dsp_tools.commands.validate_data.api_clients import OntologyClient
from dsp_tools.commands.validate_data.models.api_responses import EnabledLicenseIris
from dsp_tools.commands.validate_data.models.api_responses import ListLookup
from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import ProjectDataFromApi
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import get_rdf_like_data
from dsp_tools.commands.validate_data.prepare_data.make_data_graph import make_data_graph
from dsp_tools.commands.validate_data.sparql.construct_shacl import construct_shapes_graphs
from dsp_tools.utils.xml_parsing.get_lookups import get_authorship_lookup
from dsp_tools.utils.xml_parsing.get_parsed_resources import get_parsed_resources
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from dsp_tools.utils.xml_parsing.parse_clean_validate_xml import parse_and_clean_xml_file


def get_info_and_parsed_resources_from_file(
    file: Path, api_url: str
) -> tuple[list[ParsedResource], str, dict[str, list[str]], list[str]]:
    root = parse_and_clean_xml_file(file)
    shortcode = root.attrib["shortcode"]
    authorship_lookup = get_authorship_lookup(root)
    permission_ids = [perm.attrib["id"] for perm in root.findall("permissions")]
    parsed_resources = get_parsed_resources(root, api_url)
    return parsed_resources, shortcode, authorship_lookup, permission_ids


def prepare_data_for_validation_from_parsed_resource(
    parsed_resources: list[ParsedResource],
    authorship_lookup: dict[str, list[str]],
    permission_ids: list[str],
    auth: AuthenticationClient,
    shortcode: str,
) -> tuple[RDFGraphs, set[str]]:
    used_iris = {x.res_type for x in parsed_resources}
    proj_info = _get_project_specific_information_from_api(auth, shortcode)
    list_lookup = _make_list_lookup(proj_info.all_lists)
    data_rdf = _make_data_graph_from_parsed_resources(parsed_resources, authorship_lookup, list_lookup)
    rdf_graphs = _create_graphs(data_rdf, shortcode, auth, proj_info, permission_ids)
    return rdf_graphs, used_iris


def _make_list_lookup(project_lists: list[OneList]) -> ListLookup:
    lookup = {}
    for li in project_lists:
        for nd in li.nodes:
            lookup[(li.list_name, nd.name)] = nd.iri
            lookup[("", nd.iri)] = nd.iri
    return ListLookup(lookup)


def _get_project_specific_information_from_api(auth: AuthenticationClient, shortcode: str) -> ProjectDataFromApi:
    list_client = ListClient(auth.server, shortcode)
    all_lists = list_client.get_lists()
    enabled_licenses = _get_license_iris(shortcode, auth)
    return ProjectDataFromApi(all_lists, enabled_licenses)


def _make_data_graph_from_parsed_resources(
    parsed_resources: list[ParsedResource], authorship_lookup: dict[str, list[str]], list_lookup: ListLookup
) -> Graph:
    rdf_like_data = get_rdf_like_data(parsed_resources, authorship_lookup, list_lookup)
    rdf_data = make_data_graph(rdf_like_data)
    return rdf_data


def _create_graphs(
    data_rdf: Graph,
    shortcode: str,
    auth: AuthenticationClient,
    proj_info: ProjectDataFromApi,
    permission_ids: list[str],
) -> RDFGraphs:
    logger.debug("Create all graphs.")
    onto_client = OntologyClient(auth.server, shortcode)
    ontologies, onto_iris = _get_project_ontos(onto_client)
    knora_ttl = onto_client.get_knora_api()
    knora_api = Graph()
    knora_api.parse(data=knora_ttl, format="ttl")
    shapes = construct_shapes_graphs(ontologies, knora_api, proj_info, permission_ids)
    api_shapes = Graph()
    api_shapes_path = importlib.resources.files("dsp_tools").joinpath("resources/validate_data/api-shapes.ttl")
    api_shapes.parse(str(api_shapes_path))
    api_card_shapes = Graph()
    api_card_path = importlib.resources.files("dsp_tools").joinpath(
        "resources/validate_data/api-shapes-resource-cardinalities.ttl"
    )
    api_card_shapes.parse(str(api_card_path))
    content_shapes = shapes.content + api_shapes
    card_shapes = shapes.cardinality + api_card_shapes
    data_rdf = _bind_prefixes_to_graph(data_rdf, onto_iris)
    ontologies = _bind_prefixes_to_graph(ontologies, onto_iris)
    card_shapes = _bind_prefixes_to_graph(card_shapes, onto_iris)
    content_shapes = _bind_prefixes_to_graph(content_shapes, onto_iris)
    knora_api = _bind_prefixes_to_graph(knora_api, onto_iris)
    return RDFGraphs(
        data=data_rdf,
        ontos=ontologies,
        cardinality_shapes=card_shapes,
        content_shapes=content_shapes,
        knora_api=knora_api,
    )


def _bind_prefixes_to_graph(g: Graph, project_ontos: list[str]) -> Graph:
    def get_one_prefix(ontology_iri: str) -> str:
        iri_split = ontology_iri.split("/")
        return iri_split[-2]

    g.bind("knora-api", "http://api.knora.org/ontology/knora-api/v2#")
    g.bind("api-shapes", "http://api.knora.org/ontology/knora-api/shapes/v2#")
    g.bind("dash", "http://datashapes.org/dash#")
    g.bind("salsah-gui", "http://api.knora.org/ontology/salsah-gui/v2#")
    for iri in project_ontos:
        g.bind(get_one_prefix(iri), f"{iri}#")
    return g


def _get_project_ontos(onto_client: OntologyClient) -> tuple[Graph, list[str]]:
    logger.debug("Get project ontologies from server.")
    all_ontos, onto_iris = onto_client.get_ontologies()
    onto_g = Graph()
    for onto in all_ontos:
        og = Graph()
        og.parse(data=onto, format="ttl")
        onto_g += og
    return onto_g, onto_iris


def _get_license_iris(shortcode: str, auth: AuthenticationClient) -> EnabledLicenseIris:
    legal_client = LegalInfoClientLive(auth.server, shortcode, auth)
    license_info = legal_client.get_licenses_of_a_project()
    iris = [x["id"] for x in license_info]
    return EnabledLicenseIris(iris)



import pytest

from dsp_tools.commands.xmlupload.prepare_xml_input.ark2iri import convert_ark_v0_to_resource_iri
from dsp_tools.error.exceptions import BaseError


def test_good() -> None:
    ark = "ark:/72163/080c-779b9990a0c3f-6e"
    iri = convert_ark_v0_to_resource_iri(ark)
    assert "http://rdfh.ch/080C/Ef9heHjPWDS7dMR_gGax2Q" == iri


def test_invalid_ark() -> None:
    with pytest.raises(
        BaseError, match=r"converting ARK 'ark:/72163/080c-779b999-0a0c3f-6e'\. The ARK seems to be invalid"
    ):
        convert_ark_v0_to_resource_iri("ark:/72163/080c-779b999-0a0c3f-6e")


def test_invalid_shortcode() -> None:
    with pytest.raises(
        BaseError, match=r"converting ARK 'ark:/72163/080X-779b9990a0c3f-6e'\. Invalid project shortcode '080X'"
    ):
        convert_ark_v0_to_resource_iri("ark:/72163/080X-779b9990a0c3f-6e")


def test_invalid_shortcode_long() -> None:
    with pytest.raises(
        BaseError, match=r"converting ARK 'ark:/72163/080c1-779b9990a0c3f-6e'\. Invalid project shortcode '080C1'"
    ):
        convert_ark_v0_to_resource_iri("ark:/72163/080c1-779b9990a0c3f-6e")


def test_invalid_salsah_id() -> None:
    with pytest.raises(
        BaseError,
        match=r"converting ARK 'ark:/72163/080c-779b99\+90a0c3f-6e'\. Invalid Salsah ID '779b99\+90a0c3f'",
    ):
        convert_ark_v0_to_resource_iri("ark:/72163/080c-779b99+90a0c3f-6e")


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from datetime import datetime
from typing import cast

from loguru import logger
from rdflib import RDF
from rdflib import BNode
from rdflib import Graph
from rdflib import URIRef
from tqdm import tqdm

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.xmlupload.make_rdf_graph.jsonld_utils import serialise_jsonld_for_value
from dsp_tools.commands.xmlupload.make_rdf_graph.make_values import make_link_value_graph
from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStash
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.error.exceptions import BaseError


def upload_stashed_resptr_props(
    upload_state: UploadState,
    con: Connection,
) -> None:
    """
    After all resources are uploaded, the stashed resptr props must be applied to their resources in DSP.
    The upload state is updated accordingly, as a side effect.

    Args:
        upload_state: the current state of the upload
        con: connection to DSP
    """
    logger.info("Upload the stashed links...")
    upload_state.pending_stash = cast(Stash, upload_state.pending_stash)
    link_value_stash = cast(LinkValueStash, upload_state.pending_stash.link_value_stash)
    progress_bar = tqdm(
        link_value_stash.res_2_stash_items.copy().items(), desc="Upload the stashed links", dynamic_ncols=True
    )
    for res_id, stash_items in progress_bar:
        res_iri = upload_state.iri_resolver.get(res_id)
        if not res_iri:
            # resource could not be uploaded to DSP, so the stash cannot be uploaded either
            # no action necessary: this resource will remain in nonapplied_resptr_props,
            # which will be handled by the caller
            continue
        logger.info(f"  Upload resptrs of resource '{res_id}'...")
        for stash_item in reversed(stash_items):
            # reversed avoids any problems caused by removing from the list we loop over at the same time
            target_iri = upload_state.iri_resolver.get(stash_item.value.value)
            if not target_iri:
                continue
            if _upload_stash_item(stash_item, res_iri, target_iri, con):
                link_value_stash.res_2_stash_items[res_id].remove(stash_item)
        # remove res_id if all stash items were uploaded
        if not link_value_stash.res_2_stash_items[res_id]:
            del link_value_stash.res_2_stash_items[res_id]


def _upload_stash_item(
    stash: LinkValueStashItem,
    res_iri: str,
    target_iri: str,
    con: Connection,
) -> bool:
    """
    Upload a single stashed link value to DSP.

    Args:
        stash: the stashed link value to upload
        res_iri: the iri of the resource
        target_iri: the iri of the target resource
        con: connection to DSP

    Returns:
        True, if the upload was successful, False otherwise
    """
    graph = _make_link_value_create_graph(stash, res_iri, target_iri)
    payload = serialise_jsonld_for_value(graph, res_iri)
    try:
        con.post(route="/v2/values", data=payload)
    except BaseError as err:
        _log_unable_to_upload_link_value(err.message, stash.res_id, stash.value.prop_iri)
        return False
    logger.debug(f'  Successfully uploaded resptr links of "{stash.value.prop_iri}"')
    return True


def _make_link_value_create_graph(
    stash: LinkValueStashItem,
    res_iri_str: str,
    target_iri: str,
) -> Graph:
    """This function creates a JSON object that can be sent as an update request to the DSP-API."""
    val_bn = BNode()
    res_iri = URIRef(res_iri_str)
    graph = make_link_value_graph(stash.value, val_bn, res_iri, URIRef(target_iri))
    graph.add((res_iri, RDF.type, URIRef(stash.res_type)))
    return graph


def _log_unable_to_upload_link_value(msg: str, res_id: str, prop_name: str) -> None:
    err_msg = f"Unable to upload the resptr prop of '{prop_name}' of resource '{res_id}'."
    print(f"{datetime.now()}:     WARNING: {err_msg} Original error message: {msg}")
    logger.error(err_msg)



from abc import ABC
from abc import abstractmethod

from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT


class DspToolsWarning(Warning, ABC):
    """Abstract base class for warnings that implement a custom showwarnings() function"""

    @classmethod
    @abstractmethod
    def showwarning(cls, message: str) -> None:
        """Functionality that should be executed when a warning of this class is emitted"""


class DspToolsUserWarning(DspToolsWarning):
    """Class for general user-facing warnings"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Print the warning, without context"""
        print(BOLD_RED + f"WARNING: {message}" + RESET_TO_DEFAULT)


class DspToolsFutureWarning(DspToolsWarning, FutureWarning):
    """Class for user-facing deprecation warnings"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Print the warning, without context"""
        print(BOLD_RED + f"DEPRECATION WARNING: {message}" + RESET_TO_DEFAULT)



from dsp_tools.error.exceptions import BaseError


class XmllibInputError(BaseError):
    """
    This error is raised if a user provided invalid input.
    It should never be raised plain, but with the dedicated util function.
    """


class XmllibFileNotFoundError(BaseError):
    """
    This error is raised if a user provided a filepath that does not exist.
    """


class XmllibInternalError(BaseError):
    """
    This error is raised if an internal error, i.e. an error on which the user has no influence, is found in the xmllib.
    """



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path
from typing import Never
from typing import assert_never
from typing import cast

import pytest
from rdflib import SH
from rdflib import BNode
from rdflib import URIRef

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import SortedProblems
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.validation import DetailBaseInfo
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.process_validation_report.get_user_validation_message import sort_user_problems
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import (
    _extract_base_info_of_resource_results,
)
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import reformat_validation_graph
from dsp_tools.commands.validate_data.shacl_cli_validator import ShaclCliValidator
from dsp_tools.commands.validate_data.validate_data import _get_validation_status
from dsp_tools.commands.validate_data.validate_data import _validate_data
from dsp_tools.commands.validate_data.validation.check_duplicate_files import check_for_duplicate_files
from dsp_tools.commands.validate_data.validation.get_validation_report import get_validation_report
from dsp_tools.utils.rdflib_constants import DASH
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from test.e2e.commands.validate_data.util import prepare_data_for_validation_from_file

# ruff: noqa: ARG001 Unused function argument


CONFIG = ValidateDataConfig(
    xml_file=Path(),
    save_graph_dir=None,
    severity=ValidationSeverity.INFO,
    ignore_duplicate_files_warning=False,
    is_on_prod_server=False,
)


@pytest.fixture(scope="module")
def authentication(creds: ServerCredentials) -> AuthenticationClient:
    auth = AuthenticationClientLive(server=creds.server, email=creds.user, password=creds.password)
    return auth


@pytest.fixture(scope="module")
def every_violation_combination_once_info(
    create_generic_project, authentication, shacl_validator: ShaclCliValidator
) -> tuple[ValidationReportGraphs, list[ParsedResource]]:
    file = Path("testdata/validate-data/generic/every_violation_combination_once.xml")
    graphs, _, parsed_resources = prepare_data_for_validation_from_file(file, authentication)
    g = get_validation_report(graphs, shacl_validator)
    return g, parsed_resources


class TestWithReportGraphs:
    """
    These tests, focus confirm that the larger steps from the validate-data function behave as expected.
    Because of this, the code flow has to be copied from the primary calling function.
    It is not possible to test these steps properly without a stack.
    """

    def test_extract_identifiers_of_resource_results(
        self, every_violation_combination_once_info: tuple[ValidationReportGraphs, list[ParsedResource]]
    ) -> None:
        report_graphs, parsed_resources = every_violation_combination_once_info
        report_and_onto = report_graphs.validation_graph + report_graphs.onto_graph
        data_and_onto = report_graphs.data_graph + report_graphs.onto_graph
        result = _extract_base_info_of_resource_results(report_and_onto, data_and_onto)
        result_sorted = sorted(result, key=lambda x: str(x.focus_node_iri))
        # To query and identify the validation properly
        # we rely on knowing whether a result contains a sh:detail (those with BNodes) and those without.
        # This is something that we do not have an influence on directly
        # as the sh:detail is created by the SHACL engine and is based on the SHACL-shape.
        # If we changed a SHACL shape this may influence whether a result does or does not have a BNode
        # and consequently how we must query for it.
        expected_info = [
            (URIRef("http://data/bitstream_no_legal_info"), SH.MinCountConstraintComponent, None, None),
            (URIRef("http://data/bitstream_no_legal_info"), SH.MinCountConstraintComponent, None, None),
            (URIRef("http://data/bitstream_no_legal_info"), SH.MinCountConstraintComponent, None, None),
            (URIRef("http://data/date_month_does_not_exist"), SH.OrConstraintComponent, None, None),
            (URIRef("http://data/date_month_does_not_exist"), SH.OrConstraintComponent, None, None),
            (URIRef("http://data/date_range_first_is_ce_second_bce"), DASH.CoExistsWithConstraintComponent, None, None),
            (URIRef("http://data/date_range_wrong_yyyy"), SH.LessThanOrEqualsConstraintComponent, None, None),
            (URIRef("http://data/empty_label"), SH.PatternConstraintComponent, None, None),
            (URIRef("http://data/geoname_not_number"), SH.PatternConstraintComponent, None, None),
            (URIRef("http://data/id_card_one"), SH.MinCountConstraintComponent, None, None),
            (URIRef("http://data/id_closed_constraint"), DASH.ClosedByTypesConstraintComponent, None, None),
            (URIRef("http://data/id_max_card"), SH.MaxCountConstraintComponent, None, None),
            (URIRef("http://data/id_missing_file_value"), SH.MinCountConstraintComponent, None, None),
            (URIRef("http://data/identical_values"), SH.SPARQLConstraintComponent, None, None),
            (URIRef("http://data/image_no_legal_info"), SH.MinCountConstraintComponent, None, None),
            (URIRef("http://data/image_no_legal_info"), SH.MinCountConstraintComponent, None, None),
            (URIRef("http://data/image_no_legal_info"), SH.MinCountConstraintComponent, None, None),
            (URIRef("http://data/inexistent_license_iri"), SH.InConstraintComponent, None, None),
            (URIRef("http://data/label_with_newline"), DASH.SingleLineConstraintComponent, None, None),
            (
                URIRef("http://data/link_target_non_existent"),
                SH.NodeConstraintComponent,
                BNode,
                SH.ClassConstraintComponent,
            ),
            (
                URIRef("http://data/link_target_wrong_class"),
                SH.NodeConstraintComponent,
                BNode,
                SH.ClassConstraintComponent,
            ),
            (
                URIRef("http://data/link_to_resource_in_db"),
                SH.NodeConstraintComponent,
                BNode,
                SH.ClassConstraintComponent,
            ),
            (
                URIRef("http://data/list_node_non_existent"),
                SH.NodeConstraintComponent,
                BNode,
                SH.InConstraintComponent,
            ),
            (URIRef("http://data/missing_seqnum"), DASH.CoExistsWithConstraintComponent, None, None),
            (URIRef("http://data/richtext_standoff_link_nonexistent"), SH.ClassConstraintComponent, None, None),
            (
                URIRef("http://data/simpletext_wrong_value_type"),
                SH.NodeConstraintComponent,
                BNode,
                SH.MinCountConstraintComponent,
            ),
            (URIRef("http://data/uri_wrong_value_type"), SH.ClassConstraintComponent, None, None),
            (URIRef("http://data/video_segment_start_larger_than_end"), SH.LessThanConstraintComponent, None, None),
            (URIRef("http://data/video_segment_wrong_bounds"), SH.MinInclusiveConstraintComponent, None, None),
            (URIRef("http://data/video_segment_wrong_bounds"), SH.MinExclusiveConstraintComponent, None, None),
        ]
        assert len(result) == len(expected_info)
        for result_info, expected in zip(result_sorted, expected_info):
            assert result_info.focus_node_iri == expected[0]
            if result_info.focus_node_iri == URIRef("http://data/video_segment_wrong_bounds"):
                assert result_info.source_constraint_component in [
                    SH.MinInclusiveConstraintComponent,
                    SH.MinExclusiveConstraintComponent,
                ]
            else:
                assert result_info.source_constraint_component == expected[1], result_info.focus_node_iri
            if expected[2] is None:
                assert not result_info.detail
            else:
                detail_base_info = result_info.detail
                assert isinstance(detail_base_info, DetailBaseInfo)
                assert isinstance(detail_base_info.detail_bn, expected[2]), result_info.focus_node_iri
                assert detail_base_info.source_constraint_component == expected[3], result_info.focus_node_iri

    def test_reformat_every_constraint_once(
        self, every_violation_combination_once_info: tuple[ValidationReportGraphs, list[ParsedResource]]
    ) -> None:
        report, parsed_resources = every_violation_combination_once_info
        assert not report.conforms
        expected_violations = [
            ("date_month_does_not_exist", ProblemType.GENERIC),
            ("date_month_does_not_exist", ProblemType.GENERIC),
            ("date_range_first_is_ce_second_bce", ProblemType.GENERIC),
            ("date_range_wrong_yyyy", ProblemType.GENERIC),
            ("empty_label", ProblemType.INPUT_REGEX),
            ("geoname_not_number", ProblemType.INPUT_REGEX),
            ("id_card_one", ProblemType.MIN_CARD),
            ("id_closed_constraint", ProblemType.NON_EXISTING_CARD),
            ("id_max_card", ProblemType.MAX_CARD),
            ("id_missing_file_value", ProblemType.FILE_VALUE_MISSING),
            ("identical_values", ProblemType.DUPLICATE_VALUE),
            ("inexistent_license_iri", ProblemType.GENERIC),
            ("label_with_newline", ProblemType.GENERIC),
            ("link_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
            ("link_target_wrong_class", ProblemType.LINK_TARGET_TYPE_MISMATCH),
            ("list_node_non_existent", ProblemType.GENERIC),
            ("missing_seqnum", ProblemType.GENERIC),
            ("richtext_standoff_link_nonexistent", ProblemType.GENERIC),
            ("simpletext_wrong_value_type", ProblemType.VALUE_TYPE_MISMATCH),
            ("uri_wrong_value_type", ProblemType.VALUE_TYPE_MISMATCH),
            ("video_segment_start_larger_than_end", ProblemType.GENERIC),
            ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for start that is less than zero
            ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for the end that is zero
        ]
        expected_warnings = [
            (None, ProblemType.FILE_DUPLICATE),
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("bitstream_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
            ("image_no_legal_info", ProblemType.GENERIC),
        ]
        expected_info = [("link_to_resource_in_db", ProblemType.INEXISTENT_LINKED_RESOURCE)]
        result = reformat_validation_graph(report)
        duplicate_files = check_for_duplicate_files(parsed_resources)
        sorted_problems = sort_user_problems(result, duplicate_files)
        alphabetically_sorted_violations = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
        alphabetically_sorted_warnings = sorted(sorted_problems.user_warnings, key=lambda x: str(x.res_id))
        alphabetically_sorted_info = sorted(sorted_problems.user_info, key=lambda x: str(x.res_id))
        assert len(sorted_problems.unique_violations) == len(expected_violations)
        assert len(sorted_problems.user_warnings) == len(expected_warnings)
        assert len(sorted_problems.user_info) == len(expected_info)
        assert not sorted_problems.unexpected_shacl_validation_components
        assert not result.unexpected_results
        for one_result, expected_e in zip(alphabetically_sorted_violations, expected_violations):
            assert one_result.res_id == expected_e[0]
            assert one_result.problem_type == expected_e[1]
        for one_result, expected_w in zip(alphabetically_sorted_warnings, expected_warnings):
            assert one_result.problem_type == expected_w[1]
            assert one_result.res_id == expected_w[0]
        for one_result, expected_i in zip(alphabetically_sorted_info, expected_info):
            assert one_result.problem_type == expected_i[1]
            assert one_result.res_id == expected_i[0]
        assert not _get_validation_status(sorted_problems, is_on_prod=True)
        assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_check_for_unknown_resource_classes(authentication) -> None:
    file = Path("testdata/validate-data/generic/unknown_classes.xml")
    graphs, used_iris, parsed_resources = prepare_data_for_validation_from_file(file, authentication)
    result = _validate_data(graphs, used_iris, parsed_resources, CONFIG)
    assert not result.no_problems
    problems = result.problems
    assert isinstance(problems, UnknownClassesInData)
    expected = {"onto:NonExisting", "unknown:ClassWithEverything", "unknownClass"}
    assert problems.unknown_classes == expected


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_content_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/content_violation.xml")
    graphs, used_iris, parsed_resources = prepare_data_for_validation_from_file(file, authentication)
    result = _validate_data(graphs, used_iris, parsed_resources, CONFIG)
    msg_end_date_larger_than_start = "The end date must be equal or later than the start date."
    expected_info_tuples = [
        (
            "comment_on_value_empty",
            "onto:testUriValue",
            "The comment on the value must be a non-empty string",
        ),
        (
            "comment_on_value_whitespace",
            "onto:testUriValue",
            "The comment on the value must be a non-empty string",
        ),
        (
            "date_end_day_does_not_exist",
            "onto:testSubDate1",
            "The entered date cannot be parsed into a valid date. It may have issues with the month and/or day number.",
        ),
        (
            "date_end_month_does_not_exist",
            "onto:testSubDate1",
            "The entered date cannot be parsed into a valid date. It may have issues with the month and/or day number.",
        ),
        (
            "date_range_first_is_ce_second_bce",
            "onto:testSubDate1",
            "The start date may not be later than the end date. Please take a look if your eras are correct.",
        ),
        ("date_range_wrong_mixed_precision", "onto:testSubDate1", msg_end_date_larger_than_start),
        ("date_range_wrong_year_full_GREGORIAN", "onto:testSubDate1", msg_end_date_larger_than_start),
        ("date_range_wrong_year_full_ISLAMIC", "onto:testSubDate1", msg_end_date_larger_than_start),
        ("date_range_wrong_year_full_JULIAN", "onto:testSubDate1", msg_end_date_larger_than_start),
        ("date_range_wrong_year_good_month_wrong", "onto:testSubDate1", msg_end_date_larger_than_start),
        ("date_range_wrong_year_month_good_day_wrong", "onto:testSubDate1", msg_end_date_larger_than_start),
        ("date_range_wrong_yyyy", "onto:testSubDate1", msg_end_date_larger_than_start),
        ("date_range_wrong_yyyy_mm", "onto:testSubDate1", msg_end_date_larger_than_start),
        (
            "date_single_month_does_not_exist",
            "onto:testSubDate1",
            "The entered date cannot be parsed into a valid date. It may have issues with the month and/or day number.",
        ),
        (
            "date_single_month_does_not_exist",
            "onto:testSubDate1",
            "The entered date cannot be parsed into a valid date. It may have issues with the month and/or day number.",
        ),
        ("empty_label", "rdfs:label", "The label must be a non-empty string without newlines."),
        ("empty_text_rich", "onto:testRichtext", "The value must be a non-empty string"),
        ("empty_text_simple", "onto:testTextarea", "The value must be a non-empty string"),
        ("geoname_not_number", "onto:testGeoname", "The value must be a valid geoname code"),
        ("label_with_newline", "rdfs:label", "The label must be a non-empty string without newlines."),
        ("link_target_non_existent", "onto:testHasLinkTo", "other"),
        ("link_target_wrong_class", "onto:testHasLinkToCardOneResource", "id_9_target"),
        (
            "list_name_attrib_empty",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "list_name_non_existent",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "list_node_non_existent",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
        (
            "richtext_standoff_link_nonexistent",
            "hasStandoffLinkTo",
            "A stand-off link must target an existing resource.",
        ),
        (
            "simple_text_with_newlines",
            "onto:testSimpleText",
            "The value must be a non-empty string without newlines.",
        ),
        ("text_only_whitespace_simple", "onto:testTextarea", "The value must be a non-empty string"),
        (
            "wrong_list_used",
            "onto:testListProp",
            (
                "A valid node from the list 'firstList' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
        ),
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.res_id == expected_info[0]
        assert one_result.prop_name == expected_info[1]
        # depending on the ProblemType, the string we want to test against is in a different parameter
        if one_result.problem_type == ProblemType.INPUT_REGEX:
            assert one_result.expected == expected_info[2]
        elif one_result.problem_type == ProblemType.GENERIC:
            assert one_result.message == expected_info[2]
        elif one_result.problem_type == ProblemType.LINK_TARGET_TYPE_MISMATCH:
            assert one_result.input_value == expected_info[2]
        elif one_result.problem_type == ProblemType.INEXISTENT_LINKED_RESOURCE:
            assert one_result.input_value == expected_info[2]
        else:
            nev: Never = cast(Never, one_result.problem_type)
            assert_never(nev)
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_cardinality_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/cardinality_violation.xml")
    graphs, used_iris, parsed_resource = prepare_data_for_validation_from_file(file, authentication)
    result = _validate_data(graphs, used_iris, parsed_resource, CONFIG)
    assert not result.no_problems
    expected_info_tuples = [
        ("id_card_one", ProblemType.MIN_CARD),
        ("id_closed_constraint", ProblemType.NON_EXISTING_CARD),
        ("id_max_card", ProblemType.MAX_CARD),
        ("id_min_card", ProblemType.MIN_CARD),
        ("super_prop_no_card", ProblemType.NON_EXISTING_CARD),
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.res_id == expected_info[0]
        assert one_result.problem_type == expected_info[1]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_value_type_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/value_type_violation.xml")
    graphs, used_iris, parsed_resource = prepare_data_for_validation_from_file(file, authentication)
    result = _validate_data(graphs, used_iris, parsed_resource, CONFIG)
    assert not result.no_problems
    expected_info_tuples = [
        ("bool_wrong_value_type", "This property requires a BooleanValue", "onto:testBoolean"),
        ("color_wrong_value_type", "This property requires a ColorValue", "onto:testColor"),
        ("date_wrong_value_type", "This property requires a DateValue", "onto:testSubDate1"),
        ("decimal_wrong_value_type", "This property requires a DecimalValue", "onto:testDecimalSimpleText"),
        ("geoname_wrong_value_type", "This property requires a GeonameValue", "onto:testGeoname"),
        ("integer_wrong_value_type", "This property requires a IntValue", "onto:testIntegerSimpleText"),
        ("is_date_should_be_simpletext", "This property requires a TextValue", "onto:testTextarea"),
        ("is_link_should_be_text", "TextValue without formatting", "onto:testTextarea"),
        ("is_text_should_be_integer", "This property requires a IntValue", "onto:testIntegerSpinbox"),
        ("link_wrong_value_type", "This property requires a LinkValue", "onto:testHasLinkTo"),
        ("list_wrong_value_type", "This property requires a ListValue", "onto:testListProp"),
        ("richtext_wrong_value_type", "TextValue with formatting", "onto:testRichtext"),
        ("simpletext_wrong_value_type", "TextValue without formatting", "onto:testTextarea"),
        ("time_wrong_value_type", "This property requires a TimeValue", "onto:testTimeValue"),
        ("uri_wrong_value_type", "This property requires a UriValue", "onto:testUriValue"),
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.problem_type == ProblemType.VALUE_TYPE_MISMATCH
        assert one_result.res_id == expected_info[0]
        assert one_result.expected == expected_info[1]
        assert one_result.prop_name == expected_info[2]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_unique_value_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/unique_value_violation.xml")
    graphs, used_iris, parsed_resource = prepare_data_for_validation_from_file(file, authentication)
    result = _validate_data(graphs, used_iris, parsed_resource, CONFIG)
    assert not result.no_problems
    expected_ids = [
        "identical_values_LinkValue",
        "identical_values_listNode",
        "identical_values_richtext",
        "identical_values_valueAsString",
        "identical_values_valueHas",
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_ids)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_id in zip(alphabetically_sorted, expected_ids):
        assert one_result.problem_type == ProblemType.DUPLICATE_VALUE
        assert one_result.res_id == expected_id
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_file_value_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/file_value_violation.xml")
    graphs, used_iris, parsed_resource = prepare_data_for_validation_from_file(file, authentication)
    result = _validate_data(graphs, used_iris, parsed_resource, CONFIG)
    assert not result.no_problems
    expected_info_violation = [
        ("authorship_with_newline", ProblemType.GENERIC),
        ("copyright_holder_with_newline", ProblemType.GENERIC),
        ("empty_copyright_holder", ProblemType.INPUT_REGEX),
        ("empty_license", ProblemType.GENERIC),
        ("id_archive_missing", ProblemType.FILE_VALUE_MISSING),
        ("id_archive_unknown", ProblemType.FILE_VALUE_MISSING),
        ("id_audio_missing", ProblemType.FILE_VALUE_MISSING),
        ("id_audio_unknown", ProblemType.FILE_VALUE_MISSING),
        ("id_document_missing", ProblemType.FILE_VALUE_MISSING),
        ("id_document_unknown", ProblemType.FILE_VALUE_MISSING),
        ("id_resource_without_representation", ProblemType.FILE_VALUE_PROHIBITED),
        ("id_still_image_missing", ProblemType.FILE_VALUE_MISSING),
        ("id_still_image_unknown", ProblemType.FILE_VALUE_MISSING),
        ("id_text_missing", ProblemType.FILE_VALUE_MISSING),
        ("id_text_unknown", ProblemType.FILE_VALUE_MISSING),
        ("id_video_missing", ProblemType.FILE_VALUE_MISSING),
        ("id_video_unknown", ProblemType.FILE_VALUE_MISSING),
        ("id_wrong_file_type", ProblemType.FILE_VALUE_MISSING),
        ("inexistent_license_iri", ProblemType.GENERIC),
        ("license_not_enabled", ProblemType.GENERIC),
        ("unknown_authorship_id", ProblemType.INPUT_REGEX),
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    alphabetically_sorted_violations = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    assert len(sorted_problems.unique_violations) == len(expected_info_violation)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    for one_result, expected_info in zip(alphabetically_sorted_violations, expected_info_violation):
        assert one_result.problem_type == expected_info[1]
        assert one_result.res_id == expected_info[0]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


@pytest.mark.usefixtures("create_generic_project")
def test_reformat_dsp_inbuilt_violation(authentication) -> None:
    file = Path("testdata/validate-data/generic/dsp_inbuilt_violation.xml")
    graphs, used_iris, parsed_resource = prepare_data_for_validation_from_file(file, authentication)
    result = _validate_data(graphs, used_iris, parsed_resource, CONFIG)
    assert not result.no_problems
    expected_info_tuples = [
        ("audio_segment_target_is_video", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("audio_segment_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("link_obj_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("missing_isPartOf", ProblemType.GENERIC),
        ("missing_seqnum", ProblemType.GENERIC),
        ("non_existent_permissions_bitstream", ProblemType.GENERIC),
        ("non_existent_permissions_resource", ProblemType.GENERIC),
        ("non_existent_permissions_value", ProblemType.GENERIC),
        ("region_invalid_geometry", ProblemType.INPUT_REGEX),
        ("region_isRegionOf_resource_does_not_exist", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("region_isRegionOf_resource_not_a_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("target_must_be_a_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("target_must_be_an_image_representation", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("video_segment_start_larger_than_end", ProblemType.GENERIC),
        ("video_segment_target_is_audio", ProblemType.LINK_TARGET_TYPE_MISMATCH),
        ("video_segment_target_non_existent", ProblemType.INEXISTENT_LINKED_RESOURCE),
        ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for start that is less than zero
        ("video_segment_wrong_bounds", ProblemType.GENERIC),  # once for the end that is zero
    ]
    sorted_problems = result.problems
    assert isinstance(sorted_problems, SortedProblems)
    assert len(sorted_problems.unique_violations) == len(expected_info_tuples)
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(sorted_problems.unique_violations, key=lambda x: str(x.res_id))
    for one_result, expected_info in zip(alphabetically_sorted, expected_info_tuples):
        assert one_result.problem_type == expected_info[1]
        assert one_result.res_id == expected_info[0]
    assert not _get_validation_status(sorted_problems, is_on_prod=True)
    assert not _get_validation_status(sorted_problems, is_on_prod=False)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from dataclasses import dataclass

from rdflib import URIRef


@dataclass
class RDFPropTypeInfo:
    knora_type: URIRef
    knora_prop: URIRef
    xsd_type: URIRef | None = None


@dataclass
class AbstractFileValue:
    value: str
    metadata: FileValueMetadata


@dataclass
class FileValueMetadata:
    license_iri: str
    copyright_holder: str
    authorships: list[str]
    permissions: str | None



import urllib.parse
from collections.abc import Iterator
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

import regex
from loguru import logger
from requests import JSONDecodeError
from requests import RequestException
from requests import Session
from requests.adapters import HTTPAdapter
from requests.adapters import Retry

from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.commands.ingest_xmlupload.upload_files.upload_failures import UploadFailure
from dsp_tools.config.logger_config import LOGGER_SAVEPATH
from dsp_tools.error.exceptions import BadCredentialsError
from dsp_tools.error.exceptions import InputError

STATUS_OK = 200
STATUS_UNAUTHORIZED = 401
STATUS_FORBIDDEN = 403
STATUS_NOT_FOUND = 404
STATUS_CONFLICT = 409
STATUS_INTERNAL_SERVER_ERROR = 500
STATUS_SERVER_UNAVAILABLE = 503


@dataclass
class BulkIngestClient:
    """Client to upload multiple files to the ingest server and monitor the ingest process."""

    dsp_ingest_url: str
    authentication_client: AuthenticationClient
    shortcode: str
    imgdir: Path = field(default=Path.cwd())
    session: Session = field(init=False)
    retrieval_failures = 0

    def __post_init__(self) -> None:
        retries = 6
        self.session = Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=0.3,
            allowed_methods=None,  # means all methods
            status_forcelist=[STATUS_INTERNAL_SERVER_ERROR, STATUS_SERVER_UNAVAILABLE],
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)
        self.session.headers["Authorization"] = f"Bearer {self.authentication_client.get_token()}"

    def upload_file(
        self,
        filepath: Path,
    ) -> UploadFailure | None:
        """
        Uploads a file to the ingest server.
        The load balancer on DSP servers currently has a timeout of 60s, so we need to use a timeout of 58s.
        See https://github.com/dasch-swiss/dsp-tools/pull/1335/files#r1882508057
        # noqa: DAR101
        # noqa: DAR201
        """
        timeout = 58
        url = self._build_url_for_bulk_ingest_ingest_route(filepath)
        headers = {"Content-Type": "application/octet-stream"}
        err_msg = f"Failed to upload '{filepath}' to '{url}'."
        try:
            logger.debug(f"REQUEST: POST to {url}, timeout: {timeout}, headers: {headers}")
            with open(self.imgdir / filepath, "rb") as binary_io:
                res = self.session.post(
                    url=url,
                    headers=headers,
                    data=binary_io,  # https://requests.readthedocs.io/en/latest/user/advanced/#streaming-uploads
                    timeout=timeout,
                )
            logger.debug(f"RESPONSE: {res.status_code}")
        except RequestException as e:
            logger.exception(err_msg)
            return UploadFailure(filepath, f"Exception of requests library: {e}")
        except OSError as e:
            err_msg = f"Cannot bulk-ingest {filepath}, because the file could not be opened/read: {e.strerror}"
            logger.error(err_msg)
            return UploadFailure(filepath, err_msg)
        if res.status_code != STATUS_OK:
            logger.error(f"{err_msg}: Response {res.status_code}: {res.text}")
            return UploadFailure(filepath, res.reason, res.status_code, res.text)
        return None

    def _build_url_for_bulk_ingest_ingest_route(self, filepath: Path) -> str:
        """
        Remove the leading slash of absolute filepaths,
        because the `/project/<shortcode>/bulk-ingest/ingest` route only accepts relative paths.
        The leading slash has to be added again in the "ingest-xmlupload" step, when applying the ingest ID.

        Args:
            filepath: filepath

        Returns:
            url
        """
        quoted = regex.sub(r"^%2F", "", urllib.parse.quote(str(filepath), safe=""))
        return f"{self.dsp_ingest_url}/projects/{self.shortcode}/bulk-ingest/ingest/{quoted}"

    def trigger_ingest_process(self) -> None:
        """Start the ingest process on the server."""
        url = f"{self.dsp_ingest_url}/projects/{self.shortcode}/bulk-ingest"
        timeout = 5
        logger.debug(f"REQUEST: POST to {url}, timeout: {timeout}")
        res = self.session.post(url, timeout=timeout)
        logger.debug(f"RESPONSE: {res.status_code}: {res.text}")
        if res.status_code in [STATUS_UNAUTHORIZED, STATUS_FORBIDDEN]:
            raise BadCredentialsError("Unauthorized to start the ingest process. Please check your credentials.")
        if res.status_code == STATUS_NOT_FOUND:
            raise InputError(
                f"No assets have been uploaded for project {self.shortcode}. "
                "Before using the 'ingest-files' command, you must upload some files with the 'upload-files' command."
            )
        if res.status_code == STATUS_CONFLICT:
            msg = f"Ingest process on the server {self.dsp_ingest_url} is already running. Wait until it completes..."
            print(msg)
            logger.info(msg)
            return
        if res.status_code in [STATUS_INTERNAL_SERVER_ERROR, STATUS_SERVER_UNAVAILABLE]:
            raise InputError("Server is unavailable. Please try again later.")

        try:
            returned_shortcode = res.json().get("id")
            failed: bool = returned_shortcode != self.shortcode
        except JSONDecodeError:
            failed = True
        if failed:
            raise InputError("Failed to trigger the ingest process. Please check the server logs, or try again later.")
        print(f"Kicked off the ingest process on the server {self.dsp_ingest_url}. Wait until it completes...")
        logger.info(f"Kicked off the ingest process on the server {self.dsp_ingest_url}. Wait until it completes...")

    def retrieve_mapping_generator(self) -> Iterator[str | bool]:
        """
        Try to retrieve the mapping CSV from the server.

        Yields:
            True if the ingest process is still running.
            False if there is a server error.
            The mapping CSV if the ingest process has completed.

        Raises:
            InputError: if there are too many server errors in a row.
        """
        url = f"{self.dsp_ingest_url}/projects/{self.shortcode}/bulk-ingest/mapping.csv"
        timeout = 5
        while True:
            logger.debug(f"REQUEST: GET to {url}, timeout: {timeout}")
            res = self.session.get(url, timeout=timeout)
            logger.debug(f"RESPONSE: {res.status_code}")
            if res.status_code == STATUS_CONFLICT:
                self.retrieval_failures = 0
                logger.info("Ingest process is still running. Wait until it completes...")
                yield True
            elif res.status_code != STATUS_OK or not res.text.startswith("original,derivative"):
                self.retrieval_failures += 1
                if self.retrieval_failures > 15:
                    raise InputError(f"There were too many server errors. Please check the logs at {LOGGER_SAVEPATH}.")
                msg = "While retrieving the mapping CSV, the server responded with an unexpected status code/content."
                logger.error(msg)
                yield False
            else:
                logger.info("Ingest process completed.")
                break
        yield res.content.decode("utf-8")



from __future__ import annotations

from typing import cast

import regex
from lxml import etree

from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permissions_parsed import XmlPermission
from dsp_tools.legacy_models.projectContext import ProjectContext


def get_authorship_lookup(root: etree._Element) -> dict[str, list[str]]:
    def get_one_author(ele: etree._Element) -> str:
        # The xsd file ensures that the body of the element contains valid non-whitespace characters
        txt = cast(str, ele.text)
        txt = regex.sub(r" +", " ", txt)
        return txt.strip()

    authorship_lookup = {}
    for auth in root.iter(tag="authorship"):
        individual_authors = [get_one_author(child) for child in auth.iterchildren()]
        authorship_lookup[auth.attrib["id"]] = individual_authors
    return authorship_lookup


def get_permissions_lookup(root: etree._Element, proj_context: ProjectContext) -> dict[str, Permissions]:
    permission_ele = list(root.iter(tag="permissions"))
    permissions = [XmlPermission(permission, proj_context) for permission in permission_ele]
    permissions_dict = {permission.permission_id: permission for permission in permissions}
    permissions_lookup = {name: perm.get_permission_instance() for name, perm in permissions_dict.items()}
    return permissions_lookup



from __future__ import annotations

from enum import Enum
from enum import auto


class Permissions(Enum):
    """
    Options of permissions for resources and values:

    - `PROJECT_SPECIFIC_PERMISSIONS`: the permissions defined on project level will be applied
    - `PUBLIC`: the resource/value is visible for everyone
    - `PRIVATE`: the resource/value is only visible for project members
    - `LIMITED_VIEW`: the resource/value is visible for everyone,
      but images are blurred/watermarked for non-project members

    Deprecated terms:

    - `OPEN`: use `PUBLIC` instead
    - `RESTRICTED`: use `PRIVATE` instead
    - `RESTRICTED_VIEW`: use `LIMITED_VIEW` instead

    Examples:
        ```python
        resource = xmllib.Resource.create_new(
            res_id="ID",
            restype=":ResourceType",
            label="label",
            permissions=xmllib.Permissions.PRIVATE,
        )
        ```
    """

    PROJECT_SPECIFIC_PERMISSIONS = ""
    PUBLIC = "public"
    PRIVATE = "private"
    LIMITED_VIEW = "limited_view"

    # Deprecated terminology
    OPEN = "open"
    RESTRICTED = "restricted"
    RESTRICTED_VIEW = "restricted-view"


class NewlineReplacement(Enum):
    """
    Options how to deal with `\\n` inside rich text values.

    - `NONE`: don't modify the rich text (`\\n` will be lost, because it is meaningless in an XML file)
    - `PARAGRAPH`: replace `Start\\nEnd` with `<p>Start</p><p>End</p>`
    - `LINEBREAK`: replace `Start\\nEnd` with `Start<br/>End`

    Examples:
        ```python
        # setting the replacement options for newlines
        resource = resource.add_richtext(
            prop_name=":propName",
            value="Start\\n\\nEnd",
            newline_replacement=xmllib.NewlineReplacement.PARAGRAPH
        )
        ```
    """

    NONE = auto()
    PARAGRAPH = auto()
    LINEBREAK = auto()


class DateFormat(Enum):
    """
    Date format options for the [`reformat_date`](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/helpers/#xmllib.helpers.reformat_date) function.

    - `YYYY_MM_DD`
    - `DD_MM_YYYY`
    - `MM_DD_YYYY`
    """  # noqa: E501

    YYYY_MM_DD = auto()
    DD_MM_YYYY = auto()
    MM_DD_YYYY = auto()


class Calendar(Enum):
    """
    Calendar options for the [`reformat_date`](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/helpers/#xmllib.helpers.reformat_date) function.

    - `GREGORIAN`
    - `JULIAN`
    - `ISLAMIC`
    """  # noqa: E501

    GREGORIAN = "GREGORIAN"
    JULIAN = "JULIAN"
    ISLAMIC = "ISLAMIC"


class Era(Enum):
    """
    Era options for the [`reformat_date`](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/helpers/#xmllib.helpers.reformat_date) function.

    - `CE`
    - `BCE`
    - `AD`
    - `BC`
    """  # noqa: E501

    CE = "CE"
    BCE = "BCE"
    AD = "AD"
    BC = "BC"



from __future__ import annotations

from dataclasses import dataclass

import pandas as pd


@dataclass
class ExcelSheet:
    excel_name: str
    sheet_name: str
    col_info: Columns
    df: pd.DataFrame


@dataclass
class Columns:
    preferred_lang: str
    list_cols: list[str]
    comment_cols: list[str]
    node_cols: list[ColumnNodes]

    def __post_init__(self) -> None:
        self.node_cols = sorted(self.node_cols, key=lambda x: x.level_num, reverse=True)


@dataclass
class ColumnNodes:
    level_num: int
    columns: list[str]



import warnings
from typing import cast

import pandas as pd
import pytest
import regex
from pandas.testing import assert_frame_equal

from dsp_tools.commands.excel2json import resources as e2j
from dsp_tools.commands.excel2json.models.input_error import ExcelFileProblem
from dsp_tools.commands.excel2json.models.input_error import ExcelSheetProblem
from dsp_tools.commands.excel2json.models.input_error import MandatorySheetsMissingProblem
from dsp_tools.commands.excel2json.models.input_error import MissingValuesProblem
from dsp_tools.commands.excel2json.models.input_error import RequiredColumnMissingProblem
from dsp_tools.commands.excel2json.models.input_error import ResourceSheetNotListedProblem
from dsp_tools.commands.excel2json.models.ontology import ResourceCardinality
from dsp_tools.commands.excel2json.resources import _check_complete_gui_order
from dsp_tools.commands.excel2json.resources import _create_all_cardinalities
from dsp_tools.commands.excel2json.resources import _make_one_cardinality


class TestCheckCompleteGuiOrder:
    def test_column_does_not_exist(self) -> None:
        df = pd.DataFrame({"property": [1, 2, 3]})
        expected_df = pd.DataFrame({"property": [1, 2, 3], "gui_order": [1, 2, 3]})
        res = _check_complete_gui_order("class_name", df)
        assert_frame_equal(res, expected_df)

    def test_na_in_row(self) -> None:
        df = pd.DataFrame({"property": [1, 2, 3], "gui_order": [1, pd.NA, pd.NA]})
        expected_msg = regex.escape(
            "In the sheet 'class_name' of the file 'resources.xlsx', "
            "some rows in the column 'gui_order' are empty.\n"
            "Values have been filled in automatically, "
            "so that the gui-order reflects the order of the properties in the file."
        )
        expected_df = pd.DataFrame({"property": [1, 2, 3], "gui_order": [1, 2, 3]})
        with pytest.warns(Warning, match=expected_msg):
            res = _check_complete_gui_order("class_name", df)
        assert_frame_equal(res, expected_df)

    def test_error(self) -> None:
        df = pd.DataFrame({"property": [1, 2, 3], "gui_order": [1, 2, "a"]})
        expected_msg = regex.escape(
            "In the sheet 'class_name' of the file 'resources.xlsx', "
            "some rows in the column 'gui_order' contain invalid characters "
            "that could not be converted to an integer.\n"
            "Values have been filled in automatically, "
            "so that the gui-order reflects the order of the properties in the file."
        )
        expected_df = pd.DataFrame({"property": [1, 2, 3], "gui_order": [1, 2, 3]})
        with pytest.warns(Warning, match=expected_msg):
            res = _check_complete_gui_order("class_name", df)
        assert_frame_equal(res, expected_df)

    def test_good(self) -> None:
        df = pd.DataFrame({"property": [1, 2, 3], "gui_order": ["1", "2", "3"]})
        expected_df = pd.DataFrame({"property": [1, 2, 3], "gui_order": [1, 2, 3]})
        with warnings.catch_warnings(record=True) as catched_warnings:
            res = _check_complete_gui_order("class_name", df)
            assert len(catched_warnings) == 0
        assert_frame_equal(res, expected_df)


class TestCreateAllCardinalities:
    def test_na_in_row(self) -> None:
        df = pd.DataFrame({"property": ["1", "2", "3"], "cardinality": ["1-n", "1", "0-n"], "gui_order": [1, 2, pd.NA]})
        expected_msg = regex.escape(
            "In the sheet 'class_name' of the file 'resources.xlsx', "
            "some rows in the column 'gui_order' are empty.\n"
            "Values have been filled in automatically, "
            "so that the gui-order reflects the order of the properties in the file."
        )
        with pytest.warns(Warning, match=expected_msg):
            res = _create_all_cardinalities("class_name", df)
        assert isinstance(res, list)
        res = sorted(res, key=lambda x: x.propname)
        one = res[0]
        assert isinstance(one, ResourceCardinality)
        assert one.propname == ":1"
        assert one.cardinality == "1-n"
        assert one.gui_order == 1
        two = res[1]
        assert isinstance(two, ResourceCardinality)
        assert two.propname == ":2"
        assert two.cardinality == "1"
        assert two.gui_order == 2
        three = res[2]
        assert isinstance(three, ResourceCardinality)
        assert three.propname == ":3"
        assert three.cardinality == "0-n"
        assert three.gui_order == 3

    def test_error(self) -> None:
        df = pd.DataFrame({"property": [1, 2, 3], "cardinality": ["1-n", "1", "0-n"], "gui_order": [1, 2, "a"]})
        expected_msg = regex.escape(
            "In the sheet 'class_name' of the file 'resources.xlsx', "
            "some rows in the column 'gui_order' contain invalid characters "
            "that could not be converted to an integer.\n"
            "Values have been filled in automatically, "
            "so that the gui-order reflects the order of the properties in the file."
        )
        with pytest.warns(Warning, match=expected_msg):
            res = _create_all_cardinalities("class_name", df)
        assert isinstance(res, list)
        res = sorted(res, key=lambda x: x.propname)
        one = res[0]
        assert isinstance(one, ResourceCardinality)
        assert one.propname == ":1"
        assert one.cardinality == "1-n"
        assert one.gui_order == 1
        two = res[1]
        assert isinstance(two, ResourceCardinality)
        assert two.propname == ":2"
        assert two.cardinality == "1"
        assert two.gui_order == 2
        three = res[2]
        assert isinstance(three, ResourceCardinality)
        assert three.propname == ":3"
        assert three.cardinality == "0-n"
        assert three.gui_order == 3

    def test_good(self) -> None:
        df = pd.DataFrame({"property": [1, 2, 3], "cardinality": ["1-n", "1", "0-n"], "gui_order": ["1", "2", "3"]})
        with warnings.catch_warnings(record=True) as catched_warnings:
            res = _create_all_cardinalities("class_name", df)
            assert isinstance(res, list)
            res = sorted(res, key=lambda x: x.propname)
            one = res[0]
            assert isinstance(one, ResourceCardinality)
            assert one.propname == ":1"
            assert one.cardinality == "1-n"
            assert one.gui_order == 1
            two = res[1]
            assert isinstance(two, ResourceCardinality)
            assert two.propname == ":2"
            assert two.cardinality == "1"
            assert two.gui_order == 2
            three = res[2]
            assert isinstance(three, ResourceCardinality)
            assert three.propname == ":3"
            assert three.cardinality == "0-n"
            assert three.gui_order == 3
            assert len(catched_warnings) == 0


def test_make_one_property() -> None:
    s = pd.Series(
        {
            "property": "1",
            "gui_order": 1,
            "cardinality": "1-n",
        }
    )
    res = _make_one_cardinality(s)
    assert isinstance(res, ResourceCardinality)
    assert res.propname == ":1"
    assert res.cardinality == "1-n"
    assert res.gui_order == 1


def test_validate_individual_class_sheets_problems() -> None:
    sheet_good = pd.DataFrame({"property": ["p1"], "cardinality": ["0-n"]})
    sheet_missing_card = pd.DataFrame({"property": ["p2", "p3"], "cardinality": [pd.NA, "1-n"]})
    sheet_missing_prop = pd.DataFrame({"property": ["p4", pd.NA], "cardinality": ["0-1", "1"]})
    test_dict = {
        "sheet_good": sheet_good,
        "sheet_missing_card": sheet_missing_card,
        "sheet_missing_prop": sheet_missing_prop,
    }
    res = e2j._validate_individual_class_sheets(test_dict)
    assert len(res) == 1
    assert isinstance(res[0], MissingValuesProblem)
    card_problem = res[0].locations[0]
    assert card_problem.sheet == "sheet_missing_card"
    assert card_problem.column == "cardinality"
    assert card_problem.row == 2
    prop_problem = res[0].locations[1]
    assert prop_problem.sheet == "sheet_missing_prop"
    assert prop_problem.column == "property"
    assert prop_problem.row == 3


def test_validate_individual_class_sheets_missing_column() -> None:
    test_dict = {"sheet_missing_col": pd.DataFrame({"property": ["p5"]})}
    res = e2j._validate_individual_class_sheets(test_dict)
    assert len(res) == 1
    assert isinstance(res[0], ExcelSheetProblem)
    assert len(res[0].problems) == 1
    assert res[0].sheet_name == "sheet_missing_col"
    col_problem = cast(RequiredColumnMissingProblem, res[0].problems[0])
    assert col_problem.columns == ["cardinality"]


def test_failing_validate_excel_file() -> None:
    test_dict = {"Frenchclasses": pd.DataFrame({})}
    res = e2j._validate_excel_file(test_dict)
    assert isinstance(res, ExcelFileProblem)
    assert len(res.problems) == 1
    missing = res.problems[0]
    assert isinstance(missing, MandatorySheetsMissingProblem)
    assert missing.existing_sheets == ["Frenchclasses"]
    assert missing.mandatory_sheet == ["classes"]


class TestValidateClassesExcelSheet:
    def test_missing_sheet(self) -> None:
        test_df = pd.DataFrame({"name": ["name"], "super": ["super"]})
        res = e2j._validate_classes_excel_sheet(test_df, {"other", "name"})
        assert isinstance(res, ExcelSheetProblem)
        assert len(res.problems) == 1
        missing_sheet = res.problems[0]
        assert isinstance(missing_sheet, ResourceSheetNotListedProblem)
        assert missing_sheet.missing_names == {"other"}

    def test_good_no_sheets(self) -> None:
        test_df = pd.DataFrame({"name": ["name"], "super": ["super"]})
        res = e2j._validate_classes_excel_sheet(test_df, set())
        assert not res

    def test_good_with_sheet(self) -> None:
        test_df = pd.DataFrame({"name": ["name"], "super": ["super"]})
        res = e2j._validate_classes_excel_sheet(test_df, {"name"})
        assert not res


if __name__ == "__main__":
    pytest.main([__file__])
